"""
Systemd services
================

This module provides low-level tools for managing `systemd`_ services.

.. _systemd: http://www.freedesktop.org/wiki/Software/systemd

"""
from fabric.api import hide, settings

from burlap.utils import run_as_root


def action(action, service):
    return run_as_root('systemctl %s %s.service' % (action, service,))


def enable(service):
    """
    Enable a service.

    ::

        burlap.enable('httpd')

    .. note:: This function is idempotent.
    """
    action('enable', service)


def disable(service):
    """
    Disable a service.

    ::

        burlap.systemd.disable('httpd')

    .. note:: This function is idempotent.
    """
    action('disable', service)


def is_running(service):
    """
    Check if a service is running.

    ::

        if burlap.systemd.is_running('httpd'):
            print("Service httpd is running!")
    """
    with settings(hide('running', 'stdout', 'stderr', 'warnings'), warn_only=True):
        return action('status', service).succeeded


def start(service):
    """
    Start a service.

    ::

        if not burlap.systemd.is_running('httpd'):
            burlap.systemd.start('httpd')

    .. note:: This function is idempotent.
    """
    action('start', service)


def stop(service):
    """
    Stop a service.

    ::

        if burlap.systemd.is_running('foo'):
            burlap.systemd.stop('foo')

    .. note:: This function is idempotent.
    """
    action('stop', service)


def restart(service):
    """
    Restart a service.

    ::

        if burlap.systemd.is_running('httpd'):
            burlap.systemd.restart('httpd')
        else:
            burlap.systemd.start('httpd')
    """
    action('restart', service)


def reload(service): # pylint: disable=redefined-builtin
    """
    Reload a service.

    ::

        burlap.systemd.reload('foo')

    .. warning::

        The service needs to support the ``reload`` operation.
    """
    action('reload', service)


def start_and_enable(service):
    """
    Start and enable a service (convenience function).

    .. note:: This function is idempotent.
    """
    start(service)
    enable(service)


def stop_and_disable(service):
    """
    Stop and disable a service (convenience function).

    .. note:: This function is idempotent.
    """
    stop(service)
    disable(service)



from burlap.locale import LocalesSatchel
from burlap.tests.base import TestCase

class LocaleTests(TestCase):

    def test_get_locale_dict(self):
        s = LocalesSatchel()

        text = '''#  File generated by update-locale
LANG="en_US"
LANGUAGE="en_US:en"
'''

        self.assertEqual(s.get_locale_dict(text), {'LANG': "en_US", 'LANGUAGE': "en_US:en"})



"""
Groups
======
"""
from fabric.api import hide#, run, settings
#from burlap.utils import run_as_root

from burlap.constants import *
from burlap import Satchel
from burlap.decorators import task

class GroupSatchel(Satchel):

    name = 'group'

    @task
    def exists(self, name):
        """
        Check if a group exists.
        """
        with self.settings(hide('running', 'stdout', 'warnings'), warn_only=True):
            return self.run('getent group %(name)s' % locals()).succeeded

    @task
    def create(self, name, gid=None):
        """
        Create a new group.

        Example::

            import burlap

            if not burlap.group.exists('admin'):
                burlap.group.create('admin')

        """
        args = []
        if gid:
            args.append('-g %s' % gid)
        args.append(name)
        args = ' '.join(args)
        self.sudo('groupadd --force %s || true' % args)

    @task
    def configure(self):
        pass

group = GroupSatchel()



"""
Run test with:

    tox -c tox-full.ini -e py37-ubuntu_18_04_64 -- -s burlap/tests/functional_tests/test_buildbot.py::BuildbotTests::test_cron_check

"""
import os
import tempfile

from fabric.contrib.files import exists

from burlap.buildbot import buildbot
from burlap.tests.functional_tests.base import TestCase

class BuildbotTests(TestCase):

    def test_cron_check(self):
        try:
            # Create a mock deployment settings directory.
            deploy_from_dir = tempfile.mkdtemp()
            print('deploy_from_dir:', deploy_from_dir)
            os.makedirs(os.path.join(deploy_from_dir, 'roles/all'))
            os.makedirs(os.path.join(deploy_from_dir, 'roles/local'))
            os.makedirs(os.path.join(deploy_from_dir, buildbot.env.src_dir))
            os.chdir(deploy_from_dir)

            buildbot.genv.ROLE = 'local'
            buildbot.genv.services = ['buildbot']
            buildbot.clear_caches()

            #TODO:fix? deprecated?
            #buildbot.env.cron_check_enabled = True
            #buildbot.env.cron_check_worker_pid_path = '/usr/local/myproject/src/buildbot/worker/twistd.pid'
            #buildbot.update_cron_check()

            #assert exists(buildbot.env.cron_check_command_path)
            #assert exists(buildbot.env.cron_check_crontab_path)

            #self.fake(components=buildbot.name)

            #buildbot.env.cron_check_enabled = False
            #buildbot.update_cron_check()

            assert not exists(buildbot.env.cron_check_command_path)
            assert not exists(buildbot.env.cron_check_crontab_path)
        finally:
            buildbot.uninstall_cron_check()

    def test_configure(self):
        try:

            # Create a mock deployment settings directory.
            deploy_from_dir = tempfile.mkdtemp()
            print('deploy_from_dir:', deploy_from_dir)
            os.makedirs(os.path.join(deploy_from_dir, 'all'))
            with open(os.path.join(deploy_from_dir, 'all', 'apt-requirements.txt'), 'w') as fout:
                fout.write('python3-dev\n')
            with open(os.path.join(deploy_from_dir, 'all', 'pip-requirements.txt'), 'w') as fout:
                fout.write('buildbot[bundle]==0.9.5\n')

            # Prevent buildbot from upgrading and rebooting the server.
            packager = buildbot.get_satchel('packager')
            packager.env.initial_upgrade = False
            #packager.env.manage_custom = False

            pip = buildbot.get_satchel('pip')
            pip.env.check_permissions = False

            print('Installing Apache packages...')
            apache = buildbot.get_satchel('apache')
            apache.install_packages()

            buildbot.genv.ROLE = 'local'
            buildbot.genv.ROLES_DIR = deploy_from_dir
            buildbot.env.user = buildbot.genv.user
            buildbot.env.group = buildbot.genv.user
            buildbot.genv.services = ['buildbot']
            buildbot.clear_caches()

            buildbot.env.cron_check_enabled = True
            buildbot.env.cron_check_worker_pid_path = '/usr/local/myproject/src/buildbot/worker/twistd.pid'
            buildbot.verbose = True
#             buildbot.configure()#TODO

#             assert exists(buildbot.env.cron_check_command_path)
#             assert exists(buildbot.env.cron_check_crontab_path)
#
#             self.fake(components=buildbot.name)
#
#             buildbot.env.cron_check_enabled = False
#             buildbot.configure()
#
#             assert not exists(buildbot.env.cron_check_command_path)
#             assert not exists(buildbot.env.cron_check_crontab_path)
        finally:
            buildbot.uninstall()



from burlap.locale import LocalesSatchel
from burlap.tests.functional_tests.base import TestCase

class LocaleTests(TestCase):

    def test_locale(self):
        ts = LocalesSatchel()
        ts.verbose = True

        locale0 = ts.cat_locale()
        print('locale0:', locale0)
        locale_dict0 = ts.get_locale_dict()
        language0 = locale_dict0.get('LANGUAGE')
        print('language0:', language0)
        lang0 = locale_dict0.get('LANG')
        print('lang0:', lang0)
        lc_all0 = locale_dict0.get('LC_ALL')
        print('lc_all0:', lc_all0)

        try:
            print('Setting custom locale...')
            ts.env.language = 'en_US.UTF-8'
            ts.env.lang = 'en_US.UTF-8'
            ts.env.lc_all = 'en_US.UTF-8'
            ts.configure()

            locale1 = ts.cat_locale()
            print('locale1:', locale1)
            locale_dict1 = ts.get_locale_dict()
            language1 = locale_dict1.get('LANGUAGE')
            print('language1:', language1)
            lang1 = locale_dict1.get('LANG')
            print('lang1:', lang1)
            lc_all1 = locale_dict1.get('LC_ALL')
            print('lc_all1:', lc_all1)
            assert language1 == ts.env.language
            assert lang1 == ts.env.lang
            assert lc_all1 == ts.env.lc_all

        finally:
            print('Resetting to default locale...')
            ts.env.language = language0
            ts.env.lang = lang0
            ts.env.lc_all = lc_all0
            ts.configure()



import re

# from fabric.colors import red, green

from burlap.constants import *
from burlap import Satchel
from burlap.decorators import task
from burlap.common import print_success, print_fail

# This is sample output from `cat /prod/mdstat` after sdb failed and was replaced,
# but not before it was re-added to raid array.
SAMPLE_TEXT_FAILED_SDB = '''
Personalities : [linear] [multipath] [raid0] [raid1] [raid6] [raid5] [raid4] [raid10]
md0 : active raid1 sda1[2]
      7810036 blocks super 1.2 [2/1] [U_]

md1 : active raid1 sda2[2]
      968948600 blocks super 1.2 [2/1] [U_]

unused devices: <none>
'''

ALL_DRIVES_IN_SYSTEM_REGEX = re.compile(r"sd[a-z]")

ALL_DRIVES_IN_ARRAY_REGEX = re.compile(r"sd[a-z]")

ALL_PARTITIONS_IN_ARRAY_REGEX = re.compile(r"([a-zA-Z0-9]+)\[[0-9]+\]")

FAILED_DRIVES_REGEX = re.compile(r"([a-zA-Z0-9]+)\[[0-9]+\]\(F\)")

EMPTY_SLOTS_REGEX = re.compile(r"\[([U_]+)\]")

class SoftwareRaidSatchel(Satchel):

    name = 'softwareraid'

    def set_defaults(self):
        self.env.hdd_replace_description = None

    @task
    def raw_status(self):
        r = self.local_renderer
        ret = r.run('cat /proc/mdstat')
        return ret

    @task
    def status(self):
        r = self.local_renderer

        all_drives_in_system = r.run('ls /dev/sd*')
        all_drives_in_system = set(ALL_DRIVES_IN_SYSTEM_REGEX.findall(all_drives_in_system))
        print('all drives in system:', all_drives_in_system)

        ret = self.raw_status()

        all_drives_in_array = set(ALL_DRIVES_IN_ARRAY_REGEX.findall(ret))
        print('all drives in array:', all_drives_in_array)

        drives_needing_to_be_readded = all_drives_in_system.difference(all_drives_in_array)
        print('drives_needing_to_be_readded:', drives_needing_to_be_readded)

        all_partitions_in_array = set(ALL_PARTITIONS_IN_ARRAY_REGEX.findall(ret))
        print('all partitions in array:', all_partitions_in_array)

        bad_drives = set(FAILED_DRIVES_REGEX.findall(ret))
        print('bad_drives:', bad_drives)

        empty_slots = [_ for _ in EMPTY_SLOTS_REGEX.findall(ret) if '_' in _]
        print('empty_slots:', empty_slots)

        if not all_drives_in_system:
            print_fail('NO DRIVES FOUND! Something is very wrong!')
        if not all_drives_in_array:
            print_fail('NO DRIVES FOUND IN ARRAY! Something is very wrong!')
        elif bad_drives:
            print_fail('RAID has degraded! Shutdown and replace drive:', bad_drives)
        elif empty_slots:
            if drives_needing_to_be_readded:
                drive_str = ', '.join(sorted(drives_needing_to_be_readded))
                print_fail((
                    'RAID has degraded, the failed drive %s has been replaced, '
                    'but it needs to be re-added.') \
                    % drive_str)

                print('''
    # copy partitioning scheme from good drive onto bad drive
    sudo sfdisk -d /dev/{good_drive} | sudo sfdisk /dev/{new_drive}

    # add drive back into raid
    sudo mdadm --manage /dev/md0 --add /dev/{new_drive}1
    sudo mdadm --manage /dev/md1 --add /dev/{new_drive}2

    # wait until drives synced
    cat /proc/mdstat

    #finally install grub on new disk
    sudo grub-install /dev/{new_drive}
                '''.format(**dict(
                    good_drive=list(all_drives_in_array)[0],
                    new_drive=list(drives_needing_to_be_readded)[0],
                )))

            else:
                print_fail('RAID has degraded, the failed drive appears to have been removed, '
                    'but no new drive has been added.')
        else:
            print_success('RAID is good.')

#         print('ret:', ret)

    @task(precursors=['package'])
    def configure(self):
        pass

softwareraid = SoftwareRaidSatchel()



import re

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

# Note, using the name "locale" doesn't allow the satchel to be imported due to a conflict with an existing variable/module.

class LocalesSatchel(Satchel):

    name = 'locales'

    def set_defaults(self):
        self.env.language = 'en_US:en' # 'en_US.UTF-8'
        self.env.lang = 'C' # 'en_US.UTF-8'
        self.env.lc_all = None # 'C' # 'en_US.UTF-8'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['locales'],
            DEBIAN: ['locales'],
        }

    @task
    def cat_locale(self):
        return self.run('cat /etc/default/locale')

    def get_locale_dict(self, text=None):
        """
        Reads /etc/default/locale and returns a dictionary representing its key pairs.
        """
        text = text or self.cat_locale()
        # Format NAME="value".
        return dict(re.findall(r'^([a-zA-Z_]+)\s*=\s*[\'\"]*([0-8a-zA-Z_\.\:\-]+)[\'\"]*', text, re.MULTILINE))

    @task(precursors=['user'])
    def configure(self):
        r = self.local_renderer

        # Locales is an odd case, because it needs to be run before most packages are installed
        # but it still needs to ensure it's own package is installed.
        self.install_packages()

        args = []
        if r.env.language:
            args.append('LANGUAGE={language}')
        if r.env.lang:
            args.append('LANG={lang}')
        if r.env.lc_all:
            args.append('LC_ALL={lc_all}')

        r.env.exports = ' '.join('export %s;' % _ for _ in args)
        r.env.lang = r.env.lang or r.env.language
        if r.env.lang:
            r.sudo('{exports} locale-gen {lang}')

        r.sudo('{exports} dpkg-reconfigure --frontend=noninteractive locales')

        r.env.update_args = ' '.join(args)
        r.sudo('{exports} update-locale {update_args}')

locales = LocalesSatchel()



from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class AvahiSatchel(ServiceSatchel):

    name = 'avahi'

    def set_defaults(self):

        self.env.daemon_name = 'avahi-daemon'

        self.env.service_commands = {
            START:{
                UBUNTU: 'service %s start' % self.env.daemon_name,
                DEBIAN: 'service %s start' % self.env.daemon_name,
            },
            STOP:{
                UBUNTU: 'service %s stop' % self.env.daemon_name,
                DEBIAN: 'service %s stop' % self.env.daemon_name,
            },
            DISABLE:{
                #UBUNTU: 'chkconfig %s off' % self.env.daemon_name,
                UBUNTU: 'systemctl disable %s.service' % self.env.daemon_name,
                DEBIAN: 'update-rc.d %s disable' % self.env.daemon_name,
            },
            ENABLE:{
                #UBUNTU: 'chkconfig %s on' % self.env.daemon_name,
                UBUNTU: 'systemctl enable %s.service' % self.env.daemon_name,
                DEBIAN: 'update-rc.d %s enable' % self.env.daemon_name,
            },
            RESTART:{
                UBUNTU: 'service %s restart' % self.env.daemon_name,
                DEBIAN: 'service %s restart' % self.env.daemon_name,
            },
            STATUS:{
                UBUNTU: 'service %s status' % self.env.daemon_name,
                DEBIAN: 'service %s status' % self.env.daemon_name,
            },
        }

    @property
    def packager_system_packages(self):
        return {
            DEBIAN: ['avahi-daemon'],
            UBUNTU: ['avahi-daemon'],
        }

    @task(precursors=['packager', 'user'])
    def configure(self):
        if self.env.enabled:
            self.enable()
            self.restart()
        else:
            self.disable()
            self.stop()

AvahiSatchel()



try:
    from commands import getstatusoutput
except ImportError:
    from subprocess import getstatusoutput

from burlap.tests.base import TestCase as _TestCase
from burlap.deploy import deploy as deploy_satchel #thumbprint, clear_fs_cache, delete_plan_data_dir

class TestCase(_TestCase):

    # These are set in conftest.py before the functional tests are run, and aren't set again,
    # so tell the parent TestCase to not clear them between individual tests.
    keep_env_keys = [
        'host_string',
        'hosts',
        'roles',
        'user',
        'key_filename',
        'disable_known_hosts',
        'abort_on_prompts',
        'always_use_pty',
    ]

    def bash(self, cmd):
        cmd = cmd.replace('"', r'\"')
        s = '/bin/bash -c "%s"' % cmd
        print('cmd:', s)
        return getstatusoutput(s)

    def thumbprint(self, components=None):
        deploy_satchel.purge()
        deploy_satchel.fake(components=components)



"""
Raspberry Pi
===================

Tools for configuring software and features specific to a Raspberry Pi.
"""
import os
import getpass

import six

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

RPI2 = 'rpi2'
RPI3 = 'rpi3'

class RaspberryPiSatchel(Satchel):
    """
    Tools for configuring the Raspberry Pi.

    Most of these were written for the Raspberry Pi 2, but now target the Raspberry Pi 3.
    """

    name = 'rpi'

    def set_defaults(self):
        self.env.i2c_enabled = False
        self.env.gpio_enabled = False
        self.env.camera_enabled = False

        # The group used for allowing access to /dev/ttyACM* serial devices.
        self.env.serial_group = 'dialout'

        self.env.hardware_version = RPI3

#         self.env.firmware_update_bin_url = \
#             'https://raw.githubusercontent.com/Hexxeh/rpi-update/master/rpi-update'

        self.env.gpu_mem = 256

        # The SD card reader and mount info.
        self.env.sd_device = '/dev/sdb'
        self.env.sd_media_mount_dir = '/dev/sdb1'
        self.env.sd_media_mount_dir2 = '/dev/sdb2'

        # Raspbian specifics.
        # Should be one of the filenames found at:
        # https://downloads.raspberrypi.org
        self.env.raspbian_image_zip = 'raspbian_lite_latest.zip'
        self.env.raspbian_download_url = 'https://downloads.raspberrypi.org/raspbian_lite_latest'
        self.env.raspbian_mount_point = '/mnt/img'
        # Should be one of the filenames found at:
        # https://github.com/dhruvvyas90/qemu-rpi-kernel
        self.env.raspbian_kernel = 'kernel-qemu-4.1.13-jessie'

        # Ubuntu specifics.
        # NOTE: found to work reliably on both RPi3 and RPi2.
        self.env.ubuntu_download_url = \
            'http://phillw.net/isos/pi2/ubuntu-minimal-16.04-server-armhf-raspberry-pi.img.xz'

        self.env.conf_os_type = RASPBIAN
        self.env.conf_os_release = JESSIE

        self.env.libvirt_boot_dir = '/var/lib/libvirt/boot'
        self.env.libvirt_images_dir = '/var/lib/libvirt/images'

#         self.env.default_hostname = 'raspberrypi'
#         self.env.default_user = 'pi'
#         self.env.default_password = 'raspberry'

    @task
    def show_firmware(self):
        r = self.local_renderer
        r.sudo('cat /boot/.firmware_revision')

    @task
    def update_firmware(self):
        r = self.local_renderer
        r.pc('Updating firmware.')

#         packager = self.packager
#         if packager == APT:
#             r.sudo('apt-get install -y binutils')
#         else:
#             raise NotImplementedError
        self.install_packages()

        # Install most recent version of rpi-update, if not present.
        r.sudo_if_missing(
            fn='/usr/bin/rpi-update',
            cmd='curl -L --output /usr/bin/rpi-update {firmware_update_bin_url} && '
                'chmod +x /usr/bin/rpi-update')

        # Update firmware.
        r.sudo("sudo rpi-update")

        # Reboot to take effect.
        self.reboot(wait=300, timeout=60)

    @task
    def fix_eth0_rename(self, hardware_addr):
        """
        A bug as of 2016.10.10 causes eth0 to be renamed to enx*.
        This renames it to eth0.

        http://raspberrypi.stackexchange.com/q/43560/29103
        """
        r = self.local_renderer
        r.env.hardware_addr = hardware_addr
        r.sudo('ln -s /dev/null /etc/udev/rules.d/80-net-name-slot.rules')
        r.append(
            text=r'SUBSYSTEM=="net", ACTION=="add", DRIVERS=="?*", '\
                r'ATTR\{address\}=="{hardware_addr}", '\
                r'ATTR\{dev_id\}=="0x0", '\
                r'ATTR\{type\}=="1", '\
                r'KERNEL=="eth*", NAME="eth0"',
            filename='/etc/udev/rules.d/70-persistent-net.rules',
            use_sudo=True,
        )

    def assume_localhost(self):
        """
        Sets connection parameters to localhost, if not set already.
        """
        if not self.genv.host_string:
            self.genv.host_string = 'localhost'
            self.genv.hosts = ['localhost']
            self.genv.user = getpass.getuser()

    @task
    def init_raspbian_disk(self, yes=0):
        """
        Downloads the latest Raspbian image and writes it to a microSD card.

        Based on the instructions from:

        https://www.raspberrypi.org/documentation/installation/installing-images/linux.md
        """
        self.assume_localhost()

        yes = int(yes)
        device_question = 'SD card present at %s? ' % self.env.sd_device
        if not yes and not six.moves.input(device_question).lower().startswith('y'):
            return

        r = self.local_renderer
        r.local_if_missing(
            fn='{raspbian_image_zip}',
            cmd='wget {raspbian_download_url} -O raspbian_lite_latest.zip')

        r.lenv.img_fn = \
            r.local("unzip -l {raspbian_image_zip} | sed -n 4p | awk '{{print $4}}'", capture=True) or '$IMG_FN'
        r.local('echo {img_fn}')
        r.local('[ ! -f {img_fn} ] && unzip {raspbian_image_zip} {img_fn} || true')
        r.lenv.img_fn = r.local('readlink -f {img_fn}', capture=True)
        r.local('echo {img_fn}')

        with self.settings(warn_only=True):
            r.sudo('[ -d "{sd_media_mount_dir}" ] && umount {sd_media_mount_dir} || true')
        with self.settings(warn_only=True):
            r.sudo('[ -d "{sd_media_mount_dir2}" ] && umount {sd_media_mount_dir2} || true')

        r.pc('Writing the image onto the card.')
        r.sudo('time dd bs=4M if={img_fn} of={sd_device}')

        # Flush all writes to disk.
        r.run('sync')

    @task
    def init_ubuntu_disk(self, yes=0):
        """
        Downloads the latest Ubuntu image and writes it to a microSD card.

        Based on the instructions from:

            https://wiki.ubuntu.com/ARM/RaspberryPi

        For recommended SD card brands, see:

            http://elinux.org/RPi_SD_cards

        Note, if you get an error like:

            Kernel panic-not syncing: VFS: unable to mount root fs

        that means the SD card is corrupted. Try re-imaging the card or use a different card.
        """
        self.assume_localhost()

        yes = int(yes)

        if not self.dryrun:
            device_question = 'SD card present at %s? ' % self.env.sd_device
            inp = six.moves.input(device_question).strip()
            print('inp:', inp)
            if not yes and inp and not inp.lower().startswith('y'):
                return

        r = self.local_renderer

        # Confirm SD card is present.
        r.local('ls {sd_device}')

        # Download image.
        r.env.ubuntu_image_fn = os.path.abspath(os.path.split(self.env.ubuntu_download_url)[-1])
        r.local('[ ! -f {ubuntu_image_fn} ] && wget {ubuntu_download_url} || true')

        # Ensure SD card is unmounted.
        with self.settings(warn_only=True):
            r.sudo('[ -d "{sd_media_mount_dir}" ] && umount {sd_media_mount_dir}')
        with self.settings(warn_only=True):
            r.sudo('[ -d "{sd_media_mount_dir2}" ] && umount {sd_media_mount_dir2}')

        r.pc('Writing the image onto the card.')
        r.sudo('xzcat {ubuntu_image_fn} | dd bs=4M of={sd_device}')

        # Flush all writes to disk.
        r.run('sync')

    #EXPERIMENTAL
    @task
    def init_raspbian_vm(self):
        """
        Creates an image for running Raspbian in a QEMU virtual machine.

        Based on the guide at:

            https://github.com/dhruvvyas90/qemu-rpi-kernel/wiki/Emulating-Jessie-image-with-4.1.x-kernel
        """

        r = self.local_renderer

        r.comment('Installing system packages.')
        r.sudo('add-apt-repository ppa:linaro-maintainers/tools')
        r.sudo('apt-get update')
        r.sudo('apt-get install libsdl-dev qemu-system')

        r.comment('Download image.')
        r.local('wget https://downloads.raspberrypi.org/raspbian_lite_latest')
        r.local('unzip raspbian_lite_latest.zip')
        #TODO:fix name?
        #TODO:resize image?

        r.comment('Find start of the Linux ext4 partition.')
        r.local(
            "parted -s 2016-03-18-raspbian-jessie-lite.img unit B print | "
            "awk '/^Number/{{p=1;next}}; p{{gsub(/[^[:digit:]]/, "", $2); print $2}}' | sed -n 2p", assign_to='START')

        r.local('mkdir -p {raspbian_mount_point}')
        r.sudo('mount -v -o offset=$START -t ext4 {raspbian_image} $MNT')

        r.comment('Comment out everything in ld.so.preload')
        r.local("sed -i 's/^/#/g' {raspbian_mount_point}/etc/ld.so.preload")

        r.comment('Comment out entries containing /dev/mmcblk in fstab.')
        r.local("sed -i '/mmcblk/ s?^?#?' /etc/fstab")

        r.sudo('umount {raspbian_mount_point}')

        r.comment('Download kernel.')
        r.local('wget https://github.com/dhruvvyas90/qemu-rpi-kernel/blob/master/{raspbian_kernel}?raw=true')
        r.local('mv {raspbian_kernel} {libvirt_images_dir}')

        r.comment('Creating libvirt machine.')
        r.local('virsh define libvirt-raspbian.xml')

        r.comment('You should now be able to boot the VM by running:')
        r.comment('')
        r.comment('    qemu-system-arm -kernel {libvirt_boot_dir}/{raspbian_kernel} '
            '-cpu arm1176 -m 256 -M versatilepb -serial stdio -append "root=/dev/sda2 rootfstype=ext4 rw" '
            '-hda {libvirt_images_dir}/{raspbian_image}')
        r.comment('')
        r.comment('Or by running virt-manager.')

    @task
    def create_raspbian_vagrant_box(self):
        """
        Creates a box for easily spinning up a virtual machine with Vagrant.

        http://unix.stackexchange.com/a/222907/16477
        https://github.com/pradels/vagrant-libvirt
        """

        r = self.local_renderer

        r.sudo('adduser --disabled-password --gecos "" vagrant')

        #vagrant user should be able to run sudo commands without a password prompt

        r.sudo('echo "vagrant ALL=(ALL) NOPASSWD:ALL" > /etc/sudoers.d/vagrant')
        r.sudo('chmod 0440 /etc/sudoers.d/vagrant')

        r.sudo('apt-get update')
        r.sudo('apt-get install -y openssh-server')

        #put ssh key from vagrant user

        r.sudo('mkdir -p /home/vagrant/.ssh')
        r.sudo('chmod 0700 /home/vagrant/.ssh')
        r.sudo('wget --no-check-certificate https://raw.github.com/mitchellh/vagrant/master/keys/vagrant.pub -O /home/vagrant/.ssh/authorized_keys')
        r.sudo('chmod 0600 /home/vagrant/.ssh/authorized_keys')
        r.sudo('chown -R vagrant /home/vagrant/.ssh')

        #open sudo vi /etc/ssh/sshd_config and change

        #PubKeyAuthentication yes
        #PermitEmptyPasswords no
        r.sudo("sed -i '/AuthorizedKeysFile/s/^#//g' /etc/ssh/sshd_config")
        #PasswordAuthentication no
        r.sudo("sed -i '/PasswordAuthentication/s/^#//g' /etc/ssh/sshd_config")
        r.sudo("sed -i 's/PasswordAuthentication yes/PasswordAuthentication no/g' /etc/ssh/sshd_config")

        #restart ssh service using

        #sudo service ssh restart

        #install additional development packages for the tools to properly compile and install
        r.sudo('apt-get upgrade')
        r.sudo('apt-get install -y gcc build-essential')
        #TODO:fix? throws dpkg: error: fgets gave an empty string from `/var/lib/dpkg/triggers/File'
        #r.sudo('apt-get install -y linux-headers-rpi')

        #do any change that you want and shutdown the VM . now , come to host machine on which guest VM is running and goto
        #the /var/lib/libvirt/images/ and choose raw image in which you did the change and copy somewhere for example /test

        r.sudo('mkdir /tmp/test')
        r.sudo('cp {libvirt_images_dir}/{raspbian_image} /tmp/test')
        r.sudo('cp {libvirt_boot_dir}/{raspbian_kernel} /tmp/test')

        #create two file metadata.json and Vagrantfile in /test do entry in metadata.json
        r.render_to_file('rpi/metadata.json', '/tmp/test/metadata.json')
        r.render_to_file('rpi/Vagrantfile', '/tmp/test/Vagrantfile')

        #convert test.img to qcow2 format using
        r.sudo('qemu-img convert -f raw -O qcow2  {libvirt_images_dir}/{raspbian_image}  {libvirt_images_dir}/{raspbian_image}.qcow2')

        #rename ubuntu.qcow2 to box.img
        r.sudo('mv {libvirt_images_dir}/{raspbian_image}.qcow2 {libvirt_images_dir}/box.img')

        #Note: currently,libvirt-vagrant support only qcow2 format. so , don't change the format just rename to box.img.
        #because it takes input with name box.img by default.
        #create box

        r.sudo('cd /tmp/test; tar cvzf custom_box.box ./metadata.json ./Vagrantfile ./{raspbian_kernel} ./box.img')

        #add box to vagrant

        #vagrant box add --name custom custom_box.box

        #go to any directory where you want to initialize vagrant and run command bellow that will create Vagrant file

        #vagrant init custom

        #start configuring vagrant VM

        #vagrant up --provider=libvirt
        #TODO:fix? Error while creating domain: Error saving the server: Call to virDomainDefineXML failed: XML error: No PCI buses available

    @property
    def packager_repositories(self):
        d = {}
        # Recommended by https://wiki.ubuntu.com/ARM/RaspberryPi
        if self.env.conf_os_type == UBUNTU:
            if self.env.hardware_version == RPI2:
                d[APT] = ['ppa:ubuntu-raspi2/ppa']
            elif self.env.hardware_version == RPI3:
                d[APT] = ['ppa:ubuntu-raspi2/ppa-rpi3']
        return d

    @property
    def packager_system_packages(self):
        UBUNTU_lst = [
            'curl', 'gcc', 'python3-dev', 'git',
            'binutils',
            'rpi-update',
            'raspi-config',
        ]
        RASPBIAN_lst = [
            'curl', 'gcc', 'python3-dev', 'git',
            'rpi-update',
            'binutils',
        ]

        if self.env.i2c_enabled:
            UBUNTU_lst.extend(['python-smbus', 'i2c-tools', 'git', 'python3-dev', 'libi2c-dev'])
            RASPBIAN_lst.extend(['python-smbus', 'i2c-tools', 'git', 'python3-dev', 'libi2c-dev'])

        return {
            UBUNTU: UBUNTU_lst,
            RASPBIAN: RASPBIAN_lst,
        }

    @task
    def test_i2c(self):
        r = self.local_renderer
        #https://learn.adafruit.com/adafruits-raspberry-pi-lesson-4-gpio-setup/configuring-i2c
        r.sudo('i2cdetect -y 1')

    @task
    def configure_hdmi(self):
        """
        Configures HDMI to support hot-plugging, so it'll work even if it wasn't
        plugged in when the Pi was originally powered up.

        Note, this does cause slightly higher power consumption, so if you don't need HDMI,
        don't bother with this.

        http://raspberrypi.stackexchange.com/a/2171/29103
        """
        r = self.local_renderer

        # use HDMI mode even if no HDMI monitor is detected
        r.enable_attr(
            filename='/boot/config.txt',
            key='hdmi_force_hotplug',
            value=1,
            use_sudo=True,
        )

        # to normal HDMI mode (Sound will be sent if supported and enabled). Without this line,
        # the Raspbmc would switch to DVI (with no audio) mode by default.
        r.enable_attr(
            filename='/boot/config.txt',
            key='hdmi_drive',
            value=2,
            use_sudo=True,
        )

    @task
    def test_camera(self):
        r = self.local_renderer
        ret = r.run('vcgencmd get_camera')
        if not self.dryrun:
            assert 'detected=1' in ret, 'Camera was not detected.'
            assert 'supported=1' in ret, 'Camera is not supported.'
            print('Camera is detected and supported!')

    @task
    def configure_camera(self):
        """
        Enables access to the camera.

            http://raspberrypi.stackexchange.com/questions/14229/how-can-i-enable-the-camera-without-using-raspi-config
            https://mike632t.wordpress.com/2014/06/26/raspberry-pi-camera-setup/

        Afterwards, test with:

            /opt/vc/bin/raspistill --nopreview --output image.jpg

        Check for compatibility with:

            vcgencmd get_camera

        which should show:

            supported=1 detected=1

        """
        #TODO:check per OS? Works on Raspbian Jessie
        r = self.local_renderer
        if self.env.camera_enabled:
            r.pc('Enabling camera.')
            #TODO:fix, doesn't work on Ubuntu, which uses commented-out values

            # Set start_x=1
            #r.sudo('if grep "start_x=0" /boot/config.txt; then sed -i "s/start_x=0/start_x=1/g" /boot/config.txt; fi')
            #r.sudo('if grep "start_x" /boot/config.txt; then true; else echo "start_x=1" >> /boot/config.txt; fi')
            r.enable_attr(
                filename='/boot/config.txt',
                key='start_x',
                value=1,
                use_sudo=True,
            )

            # Set gpu_mem=128
#             r.sudo('if grep "gpu_mem" /boot/config.txt; then true; else echo "gpu_mem=128" >> /boot/config.txt; fi')
            r.enable_attr(
                filename='/boot/config.txt',
                key='gpu_mem',
                value=r.env.gpu_mem,
                use_sudo=True,
            )

            # Compile the Raspberry Pi binaries.
            #https://github.com/raspberrypi/userland
            r.run('cd ~; git clone https://github.com/raspberrypi/userland.git; cd userland; ./buildme')
            r.run('touch ~/.bash_aliases')
            #r.run("echo 'PATH=$PATH:/opt/vc/bin\nexport PATH' >> ~/.bash_aliases")
            r.append(r'PATH=$PATH:/opt/vc/bin\nexport PATH', '~/.bash_aliases')
            #r.run("echo 'LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/opt/vc/lib\nexport LD_LIBRARY_PATH' >> ~/.bash_aliases")
            r.append(r'LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/opt/vc/lib\nexport LD_LIBRARY_PATH', '~/.bash_aliases')
            r.run('source ~/.bashrc')
            r.sudo('ldconfig')

            # Allow our user to access the video device.
            r.sudo("echo 'SUBSYSTEM==\"vchiq\",GROUP=\"video\",MODE=\"0660\"' > /etc/udev/rules.d/10-vchiq-permissions.rules")
            r.sudo("usermod -a -G video {user}")

            r.reboot(wait=300, timeout=60)

            self.test_camera()

        else:
            r.disable_attr(
                filename='/boot/config.txt',
                key='start_x',
                use_sudo=True,
            )
            r.disable_attr(
                filename='/boot/config.txt',
                key='gpu_mem',
                use_sudo=True,
            )
            r.reboot(wait=300, timeout=60)

    @task
    def fix_lsmod_for_pi3(self):
        """
        Some images purporting to support both the Pi2 and Pi3 use the wrong kernel modules.
        """
        r = self.local_renderer
        r.env.rpi2_conf = '/etc/modules-load.d/rpi2.conf'
        r.sudo("sed '/bcm2808_rng/d' {rpi2_conf}")
        r.sudo("echo bcm2835_rng >> {rpi2_conf}")

    @task
    def configure_gpio(self):
        #TODO:check per OS? Works on Raspbian Jessie and Ubuntu 16.
        r = self.local_renderer
        if self.env.gpio_enabled:
            r.pc('Enabling GPIO.')

            # assumes init_project ran first
            r.sudo("usermod -a -G gpio {user}")
            #sudo chown root:gpio /sys/class/gpio/unexport /sys/class/gpio/export
            #sudo chmod 220 /sys/class/gpio/unexport /sys/class/gpio/export

            # Make GPIO accessible to non-root users.
            #Obsolete in Ubuntu 16?
#             r.sudo("echo 'SUBSYSTEM==\"gpio*\", PROGRAM=\"/bin/sh -c '"
#                 "chown -R root:gpio /sys/class/gpio && chmod -R 770 /sys/class/gpio; "
#                 "chown -R root:gpio /sys/devices/virtual/gpio && "
#                 "chmod -R 770 /sys/devices/virtual/gpio'\"' > /etc/udev/rules.d/99-com.rules")
        else:
            pass

    @task
    def configure_sound(self):
        r = self.local_renderer
        r.sudo('usermod -G audio {user}')
        r.enable_attr(
            filename='/boot/config.txt',
            key='dtparam=audio',
            value='on',
            use_sudo=True,
        )

    @task
    def test_sound(self):
        r = self.local_renderer
        ret = r.sudo('aplay /usr/share/sounds/alsa/Front_Center.wav')
        assert 'error' not in ret.lower()

    @task
    def test_gpio(self, pin=20):
        r = self.local_renderer
        r.env.gpio_dir = '/sys/class/gpio'
        r.env.gpio_pin = pin
        r.run('cd {gpio_dir}; echo {gpio_pin} > export')
        r.run('cd {gpio_dir}; echo out > gpio{gpio_pin}/direction')
        r.run('cd {gpio_dir}; echo 1 > gpio{gpio_pin}/value')
        r.run('sleep 1')
        r.run('cd {gpio_dir}; echo 0 > gpio{gpio_pin}/value')
        r.run('cd {gpio_dir}; echo {gpio_pin} > unexport')

    @task
    def configure_serial(self):
        r = self.local_renderer
        r.sudo('usermod -a -G {serial_group} {user}')

    @task
    def configure_i2c(self):
        #TODO:fix? causes RPi3 to become unbootable?
        r = self.local_renderer
        if self.env.i2c_enabled:
            r.pc('Enabling I2C.')

            #r.sudo('apt-get install --yes python-smbus i2c-tools git python3-dev')

#             r.sudo("sh -c 'echo \"i2c-bcm2708\" >> /etc/modules'")
#             r.sudo("sh -c 'echo \"i2c-dev\" >> /etc/modules'")
#             r.sudo("sh -c 'echo \"dtparam=i2c1=on\" >> /boot/config.txt'")
#             r.sudo("sh -c 'echo \"dtparam=i2c_arm=on\" >> /boot/config.txt'")

            # Allow non-root users to access I2C.
            # http://quick2wire.com/non-root-access-to-spi-on-the-pi/
            # https://github.com/fivdi/i2c-bus/blob/master/doc/raspberry-pi-i2c.md
            # https://blogs.ncl.ac.uk/francisfranklin/2014/03/23/using-i2c-with-the-raspberry-pi-step-1-modules-and-packages/
            r.sudo('groupadd -f --system spi')
            r.sudo('adduser {user} spi')
            r.sudo('adduser {user} i2c')
            r.append(text='SUBSYSTEM=="spidev", GROUP="spi"', filename='/etc/udev/rules.d/90-spi.rules', use_sudo=True)
            r.append(text='SUBSYSTEM=="i2c-dev", MODE="0666"', filename='/etc/udev/rules.d/99-i2c.rules', use_sudo=True)
            r.append(text='KERNEL=="i2c-[0-7]",MODE="0666"', filename='/etc/udev/rules.d/90-i2c.rules', use_sudo=True)

            r.append(text='i2c-bcm2708', filename='/etc/modules', use_sudo=True)
            r.append(text='i2c-dev', filename='/etc/modules', use_sudo=True)
            r.append(text='dtparam=i2c1=on', filename='/boot/config.txt', use_sudo=True)
            r.append(text='dtparam=i2c_arm=on', filename='/boot/config.txt', use_sudo=True)

            r.reboot(wait=300, timeout=60)

            # If I2C is working, running this should show addresses in use.
            ret = r.sudo('i2cdetect -y 1')
            if not self.dryrun:
                assert ret, 'I2C configuration failed!'
        else:
            pass

    @task(precursors=['packager', 'user', 'timezone', 'avahi', 'nm', 'ntpclient', 'sshnice'])
    def configure(self):
        self.update_firmware()
        self.configure_i2c()
        self.configure_camera()
        self.configure_gpio()
        self.configure_serial()
        self.configure_sound()

RaspberryPiSatchel()



"""
RPM packages
============

This module provides tools to manage CentOS/RHEL/SL/Fedora packages
and repositories.

"""
from fabric.api import hide, run, settings

import six

from burlap.utils import run_as_root


MANAGER = 'yum -y --color=never'


def update(kernel=False):
    """
    Upgrade all packages, skip obsoletes if ``obsoletes=0`` in ``yum.conf``.

    Exclude *kernel* upgrades by default.
    """
    manager = MANAGER
    cmds = {'yum -y --color=never': {False: '--exclude=kernel* update', True: 'update'}}
    cmd = cmds[manager][kernel]
    run_as_root("%(manager)s %(cmd)s" % locals())


def upgrade(kernel=False):
    """
    Upgrade all packages, including obsoletes.

    Exclude *kernel* upgrades by default.
    """
    manager = MANAGER
    cmds = {'yum -y --color=never': {False: '--exclude=kernel* upgrade', True: 'upgrade'}}
    cmd = cmds[manager][kernel]
    run_as_root("%(manager)s %(cmd)s" % locals())


def groupupdate(group, options=None):
    """
    Update an existing software group, skip obsoletes if ``obsoletes=1`` in ``yum.conf``.

    Extra *options* may be passed to ``yum`` if necessary.
    """
    manager = MANAGER
    if options is None:
        options = []
    elif isinstance(options, str):
        options = [options]
    options = " ".join(options)
    run_as_root('%(manager)s %(options)s groupupdate "%(group)s"' % locals())


def is_installed(pkg_name):
    """
    Check if an RPM package is installed.
    """
    manager = MANAGER
    with settings(hide('running', 'stdout', 'stderr', 'warnings'), warn_only=True):
        res = run("rpm --query %(pkg_name)s" % locals())
        if res.succeeded:
            return True
        return False


def install(packages, repos=None, yes=None, options=None):
    """
    Install one or more RPM packages.

    Extra *repos* may be passed to ``yum`` to enable extra repositories at install time.

    Extra *yes* may be passed to ``yum`` to validate license if necessary.

    Extra *options* may be passed to ``yum`` if necessary
    (e.g. ``['--nogpgcheck', '--exclude=package']``).

    ::

        import burlap

        # Install a single package, in an alternative install root
        burlap.rpm.install('emacs', options='--installroot=/my/new/location')

        # Install multiple packages silently
        burlap.rpm.install([
            'unzip',
            'nano'
        ], '--quiet')

    """
    manager = MANAGER
    if options is None:
        options = []
    elif isinstance(options, six.string_types):
        options = [options]
    if not isinstance(packages, six.string_types):
        packages = " ".join(packages)
    if repos:
        for repo in repos:
            options.append('--enablerepo=%(repo)s' % locals())
    options = " ".join(options)
    if isinstance(yes, str):
        run_as_root('yes %(yes)s | %(manager)s %(options)s install %(packages)s' % locals())
    else:
        run_as_root('%(manager)s %(options)s install %(packages)s' % locals())


def groupinstall(group, options=None):
    """
    Install a group of packages.

    You can use ``yum grouplist`` to get the list of groups.

    Extra *options* may be passed to ``yum`` if necessary like
    (e.g. ``['--nogpgcheck', '--exclude=package']``).

    ::

        import burlap

        # Install development packages
        burlap.rpm.groupinstall('Development tools')

    """
    manager = MANAGER
    if options is None:
        options = []
    elif isinstance(options, str):
        options = [options]
    options = " ".join(options)
    run_as_root('%(manager)s %(options)s groupinstall "%(group)s"' % locals(), pty=False)


def uninstall(packages, options=None):
    """
    Remove one or more packages.

    Extra *options* may be passed to ``yum`` if necessary.

    """
    manager = MANAGER
    if options is None:
        options = []
    elif isinstance(options, six.string_types):
        options = [options]
    if not isinstance(packages, six.string_types):
        packages = " ".join(packages)
    options = " ".join(options)
    run_as_root('%(manager)s %(options)s remove %(packages)s' % locals())


def groupuninstall(group, options=None):
    """
    Remove an existing software group.

    Extra *options* may be passed to ``yum`` if necessary.

    """
    manager = MANAGER
    if options is None:
        options = []
    elif isinstance(options, str):
        options = [options]
    options = " ".join(options)
    run_as_root('%(manager)s %(options)s groupremove "%(group)s"' % locals())


def repolist(status='', media=None):
    """
    Get the list of ``yum`` repositories.

    Returns enabled repositories by default. Extra *status* may be passed
    to list disabled repositories if necessary.

    Media and debug repositories are kept disabled, except if you pass *media*.

    ::

        import burlap

        # Install a package that may be included in disabled repositories
        burlap.rpm.install('vim', burlap.rpm.repolist('disabled'))

    """
    manager = MANAGER
    with settings(hide('running', 'stdout')):
        if media:
            repos = run_as_root("%(manager)s repolist %(status)s | sed '$d' | sed -n '/repo id/,$p'" % locals())
        else:
            repos = run_as_root("%(manager)s repolist %(status)s | sed '/Media\\|Debug/d' | sed '$d' | sed -n '/repo id/,$p'" % locals())
        return [line.split(' ')[0] for line in repos.splitlines()[1:]]



"""
Utilities
=========
"""
from pipes import quote
import os
import posixpath
import hashlib

import six

from fabric.api import env, hide, run, sudo


def run_as_root(command, *args, **kwargs):
    """
    Run a remote command as the root user.

    When connecting as root to the remote system, this will use Fabric's
    ``run`` function. In other cases, it will use ``sudo``.
    """
    if env.user == 'root':
        func = run
    else:
        func = sudo
    return func(command, *args, **kwargs)


def get_cwd(local=False):

    from fabric.api import local as local_run

    with hide('running', 'stdout'):
        if local:
            return local_run('pwd', capture=True)
        return run('pwd')


def abspath(path, local=False):

    path_mod = os.path if local else posixpath

    if not path_mod.isabs(path):
        cwd = get_cwd(local=local)
        path = path_mod.join(cwd, path)

    return path_mod.normpath(path)


def download(url, retry=10):
#     from burlap.require.curl import command as require_curl
#     require_curl()
    run('curl --silent --retry %s -O %s' % (retry, url))


def read_file(path):
    with hide('running', 'stdout'):
        return run('cat {0}'.format(quote(path)))


def read_lines(path):
    return read_file(path).splitlines()


_oct = oct
def oct(v, **kwargs): # pylint: disable=redefined-builtin
    """
    A backwards compatible version of oct() that works with Python2.7 and Python3.
    """
    v = str(v)
    if six.PY2:
        if v.startswith('0o'):
            v = '0' + v[2:]
    else:
        if not v.starswith('0o'):
            assert v[0] == '0'
            v = '0o' + v[1:]
    return eval('_oct(%s, **kwargs)' % v) # pylint: disable=eval-used


def get_file_hash(fin, block_size=2**20):
    """
    Iteratively builds a file hash without loading the entire file into memory.
    Designed to process an arbitrary binary file.
    """
    if isinstance(fin, six.string_types):
        fin = open(fin) # pylint: disable=consider-using-with
    h = hashlib.sha512()
    while True:
        data = fin.read(block_size)
        if not data:
            break
        try:
            h.update(data)
        except TypeError:
            # Fixes Python3 error "TypeError: Unicode-objects must be encoded before hashing".
            h.update(data.encode('utf-8'))
    return h.hexdigest()



"""
Files and directories
=====================
"""
from pipes import quote
import os
from tempfile import mkstemp
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse
import hashlib

import six

from fabric.api import run as _run, sudo as _sudo, abort, warn
from fabric.api import hide
from fabric.contrib.files import upload_template as _upload_template
from fabric.contrib.files import exists

from burlap.decorators import task
from burlap.utils import run_as_root
from burlap import ContainerSatchel

BLOCKSIZE = 2 ** 20 # 1MB

class watch:
    """
    Context manager to watch for changes to the contents of some files.

    The *filenames* argument can be either a string (single filename)
    or a list (multiple filenames).

    You can read the *changed* attribute at the end of the block to
    check if the contents of any of the watched files has changed.

    You can also provide a *callback* that will be called at the end of
    the block if the contents of any of the watched files has changed.

    Example using an explicit check::

        from fabric.contrib.files import comment, uncomment

        from burlap.files import watch
        from burlap.services import restart

        # Edit configuration file
        with watch('/etc/daemon.conf') as config:
            uncomment('/etc/daemon.conf', 'someoption')
            comment('/etc/daemon.conf', 'otheroption')

        # Restart daemon if needed
        if config.changed:
            restart('daemon')

    Same example using a callback::

        from functools import partial

        from fabric.contrib.files import comment, uncomment

        from burlap.files import watch
        from burlap.services import restart

        with watch('/etc/daemon.conf', callback=partial(restart, 'daemon')):
            uncomment('/etc/daemon.conf', 'someoption')
            comment('/etc/daemon.conf', 'otheroption')

    """

    def __init__(self, filenames, callback=None, use_sudo=False):
        if isinstance(filenames, six.string_types):
            self.filenames = [filenames]
        else:
            self.filenames = filenames
        self.callback = callback
        self.use_sudo = use_sudo
        self.digest = dict()
        self.changed = False

    def __enter__(self):
        with self.settings(hide('warnings')):
            for filename in self.filenames:
                self.digest[filename] = file.md5sum(filename, self.use_sudo)
        return self

    def __exit__(self, type, value, tb): # pylint: disable=redefined-builtin
        for filename in self.filenames:
            if self.md5sum(filename, self.use_sudo) != self.digest[filename]:
                self.changed = True
                break
        if self.changed and self.callback:
            self.callback()

class FileSatchel(ContainerSatchel):
    """
    Wrapper around common file operations.

    Note, most readonly actions ignore the global dryrun setting.
    """

    name = 'file'

    @task
    def is_file(self, path, use_sudo=False):
        """
        Check if a path exists, and is a file.
        """
        if self.is_local and not use_sudo:
            return os.path.isfile(path)
        func = use_sudo and _sudo or _run
        with self.settings(hide('running', 'warnings'), warn_only=True):
            return func('[ -f "%(path)s" ]' % locals()).succeeded

    @task
    def is_dir(self, path, use_sudo=False):
        """
        Check if a path exists, and is a directory.
        """
        if self.is_local and not use_sudo:
            return os.path.isdir(path)
        func = use_sudo and _sudo or _run
        with self.settings(hide('running', 'warnings'), warn_only=True):
            return func('[ -d "%(path)s" ]' % locals()).succeeded

    @task
    def is_link(self, path, use_sudo=False):
        """
        Check if a path exists, and is a symbolic link.
        """
        func = use_sudo and _sudo or _run
        with self.settings(hide('running', 'warnings'), warn_only=True):
            return func('[ -L "%(path)s" ]' % locals()).succeeded

    @task
    def get_owner(self, path, use_sudo=False):
        """
        Get the owner name of a file or directory.
        """
        func = use_sudo and run_as_root or self.run
        # I'd prefer to use quiet=True, but that's not supported with older
        # versions of Fabric.
        with self.settings(hide('running', 'stdout'), warn_only=True):
            result = func('stat -c %%U "%(path)s"' % locals())
            if result.failed and 'stat: illegal option' in result:
                # Try the BSD version of stat
                return func('stat -f %%Su "%(path)s"' % locals())
            return result

    @task
    def get_group(self, path, use_sudo=False):
        """
        Get the group name of a file or directory.
        """
        func = use_sudo and run_as_root or self.run
        # I'd prefer to use quiet=True, but that's not supported with older
        # versions of Fabric.
        with self.settings(hide('running', 'stdout'), warn_only=True):
            result = func('stat -c %%G "%(path)s"' % locals())
            if result.failed and 'stat: illegal option' in result:
                # Try the BSD version of stat
                return func('stat -f %%Sg "%(path)s"' % locals())
            return result

    def get_mode(self, path, use_sudo=False):
        """
        Get the mode (permissions) of a file or directory.

        Returns a string such as ``'0755'``, representing permissions as
        an octal number.
        """
        func = use_sudo and run_as_root or self.run
        # I'd prefer to use quiet=True, but that's not supported with older
        # versions of Fabric.
        with self.settings(hide('running', 'stdout'), warn_only=True):
            result = func('stat -c %%a "%(path)s"' % locals())
            if result.failed and 'stat: illegal option' in result:
                # Try the BSD version of stat
                return func('stat -f %%Op "%(path)s"|cut -c 4-6' % locals())
            return result

    @task
    def umask(self, use_sudo=False):
        """
        Get the user's umask.

        Returns a string such as ``'0002'``, representing the user's umask
        as an octal number.

        If `use_sudo` is `True`, this function returns root's umask.
        """
        func = use_sudo and run_as_root or self.run
        return func('umask')

    @task
    def upload_template(self, filename, destination, context=None, use_jinja=False,
                        template_dir=None, use_sudo=False, backup=True,
                        mirror_local_mode=False, mode=None,
                        mkdir=False, chown=False, user=None):
        """
        Upload a template file.

        This is a wrapper around :func:`fabric.contrib.files.upload_template`
        that adds some extra parameters.

        If ``mkdir`` is True, then the remote directory will be created, as
        the current user or as ``user`` if specified.

        If ``chown`` is True, then it will ensure that the current user (or
        ``user`` if specified) is the owner of the remote file.
        """

        if mkdir:
            remote_dir = os.path.dirname(destination)
            if use_sudo:
                self.sudo('mkdir -p %s' % quote(remote_dir), user=user)
            else:
                self.run('mkdir -p %s' % quote(remote_dir))

        if not self.dryrun:
            _upload_template(
                filename=filename,
                destination=destination,
                context=context,
                use_jinja=use_jinja,
                template_dir=template_dir,
                use_sudo=use_sudo,
                backup=backup,
                mirror_local_mode=mirror_local_mode,
                mode=mode,
            )

        if chown:
            if user is None:
                user = self.genv.user
            run_as_root('chown %s: %s' % (user, quote(destination)))

    @task
    def md5sum(self, filename, use_sudo=False):
        """
        Compute the MD5 sum of a file.
        """
        func = use_sudo and run_as_root or self.run
        with self.settings(hide('running', 'stdout', 'stderr', 'warnings'), warn_only=True):
            # Linux (LSB)
            if exists(u'/usr/bin/md5sum'):
                res = func(u'/usr/bin/md5sum %(filename)s' % locals())
            # BSD / OS X
            elif exists(u'/sbin/md5'):
                res = func(u'/sbin/md5 -r %(filename)s' % locals())
            # SmartOS Joyent build
            elif exists(u'/opt/local/gnu/bin/md5sum'):
                res = func(u'/opt/local/gnu/bin/md5sum %(filename)s' % locals())
            # SmartOS Joyent build
            # (the former doesn't exist, at least on joyent_20130222T000747Z)
            elif exists(u'/opt/local/bin/md5sum'):
                res = func(u'/opt/local/bin/md5sum %(filename)s' % locals())
            # Try to find ``md5sum`` or ``md5`` on ``$PATH`` or abort
            else:
                md5sum = func(u'which md5sum')
                md5 = func(u'which md5')
                if exists(md5sum):
                    res = func('%(md5sum)s %(filename)s' % locals())
                elif exists(md5):
                    res = func('%(md5)s %(filename)s' % locals())
                else:
                    abort('No MD5 utility was found on this system.')

        if res.succeeded:
            _md5sum = res
        else:
            warn(res)
            _md5sum = None

        if isinstance(_md5sum, six.string_types):
            _md5sum = _md5sum.strip().split('\n')[-1].split()[0]

        return _md5sum

    @task
    def uncommented_lines(self, filename, use_sudo=False):
        """
        Get the lines of a remote file, ignoring empty or commented ones
        """
        func = run_as_root if use_sudo else self.run
        res = func('cat %s' % quote(filename), quiet=True)
        if res.succeeded:
            return [line for line in res.splitlines() if line and not line.startswith('#')]
        return []


    def getmtime(self, path, use_sudo=False):
        """
        Return the time of last modification of path.
        The return value is a number giving the number of seconds since the epoch

        Same as :py:func:`os.path.getmtime()`
        """
        func = use_sudo and run_as_root or self.run
        with self.settings(hide('running', 'stdout')):
            return int(func('stat -c %%Y "%(path)s" ' % locals()).strip())

    @task
    def copy(self, source, destination, recursive=False, use_sudo=False):
        """
        Copy a file or directory
        """
        func = use_sudo and run_as_root or self.run
        options = '-r ' if recursive else ''
        func('/bin/cp {0}{1} {2}'.format(options, quote(source), quote(destination)))

    @task
    def move(self, source, destination, use_sudo=False):
        """
        Move a file or directory
        """
        func = use_sudo and run_as_root or self.run
        func('/bin/mv {0} {1}'.format(quote(source), quote(destination)))

    @task
    def symlink(self, source, destination, use_sudo=False):
        """
        Create a symbolic link to a file or directory
        """
        func = use_sudo and run_as_root or self.run
        func('/bin/ln -s {0} {1}'.format(quote(source), quote(destination)))

    @task
    def remove(self, path, recursive=False, use_sudo=False):
        """
        Remove a file or directory
        """
        func = use_sudo and run_as_root or self.run
        options = '-r ' if recursive else ''
        func('/bin/rm {0}{1}'.format(options, quote(path)))

    @task
    def upload(self, src, dst=None):
        dst = self.put(local_path=src, remote_path=dst)
        print('Uploaded to %s' % (dst,))

    @task
    def download(self, src, dst=None):
        dst = self.get(local_path=dst, remote_path=src)
        print('Downloaded to %s' % (dst,))

    def require(self, path=None, contents=None, source=None, url=None, md5=None,
         use_sudo=False, owner=None, group='', mode=None, verify_remote=True,
         temp_dir='/tmp'):
        """
        Require a file to exist and have specific contents and properties.

        You can provide either:

        - *contents*: the required contents of the file::

            from fabtools import require

            require.file('/tmp/hello.txt', contents='Hello, world')

        - *source*: the local path of a file to upload::

            from fabtools import require

            require.file('/tmp/hello.txt', source='files/hello.txt')

        - *url*: the URL of a file to download (*path* is then optional)::

            from fabric.api import cd
            from fabtools import require

            with cd('tmp'):
                require.file(url='http://example.com/files/hello.txt')

        If *verify_remote* is ``True`` (the default), then an MD5 comparison
        will be used to check whether the remote file is the same as the
        source. If this is ``False``, the file will be assumed to be the
        same if it is present. This is useful for very large files, where
        generating an MD5 sum may take a while.

        When providing either the *contents* or the *source* parameter, Fabric's
        ``put`` function will be used to upload the file to the remote host.
        When ``use_sudo`` is ``True``, the file will first be uploaded to a temporary
        directory, then moved to its final location. The default temporary
        directory is ``/tmp``, but can be overridden with the *temp_dir* parameter.
        If *temp_dir* is an empty string, then the user's home directory will
        be used.

        If `use_sudo` is `True`, then the remote file will be owned by root,
        and its mode will reflect root's default *umask*. The optional *owner*,
        *group* and *mode* parameters can be used to override these properties.

        .. note:: This function can be accessed directly from the
                  ``fabtools.require`` module for convenience.

        """
        func = use_sudo and run_as_root or self.run

        # 1) Only a path is given
        if path and not (contents or source or url):
            assert path
            if not self.is_file(path):
                func('touch "%(path)s"' % locals())

        # 2) A URL is specified (path is optional)
        elif url:
            if not path:
                path = os.path.basename(urlparse(url).path)

            if not self.is_file(path) or md5 and self.md5sum(path) != md5:
                func('wget --progress=dot:mega "%(url)s" -O "%(path)s"' % locals())

        # 3) A local filename, or a content string, is specified
        else:
            if source:
                assert not contents
                t = None
            else:
                fd, source = mkstemp()
                t = os.fdopen(fd, 'w')
                t.write(contents)
                t.close()

            if verify_remote:
                # Avoid reading the whole file into memory at once
                digest = hashlib.md5()
                with open(source, 'rb') as f:
                    while True:
                        d = f.read(BLOCKSIZE)
                        if not d:
                            break
                        digest.update(d)
            else:
                digest = None

            if (not self.is_file(path, use_sudo=use_sudo) or
                    (verify_remote and
                        self.md5sum(path, use_sudo=use_sudo) != digest.hexdigest())):
                with self.settings(hide('running')):
                    self.put(local_path=source, remote_path=path, use_sudo=use_sudo, temp_dir=temp_dir)

            if t is not None:
                os.unlink(source)

        # Ensure correct owner
        if use_sudo and owner is None:
            owner = 'root'
        if (owner and self.get_owner(path, use_sudo) != owner) or \
           (group and self.get_group(path, use_sudo) != group):
            func('chown %(owner)s:%(group)s "%(path)s"' % locals())

        # Ensure correct mode
        if use_sudo and mode is None:
            mode = oct(0o666 & ~int(self.umask(use_sudo=True), base=8))
        if mode and self.get_mode(path, use_sudo) != mode:
            func('chmod %(mode)s "%(path)s"' % locals())

file = FileSatchel() # pylint: disable=redefined-builtin



import os
import hashlib
import pickle
from copy import deepcopy
try:
    from commands import getoutput
except ImportError:
    from subprocess import getoutput

import six


class BaseTracker:

    def __init__(self, action=None):
        if action:
            assert callable(action), 'Action %s is not callable.' % action
        self.action = action

    def natural_key(self):
        """
        This is a string or sequence that uniquely identifies this tracker.
        """
        raise NotImplementedError

    def get_natural_key_hash(self):
        m = hashlib.md5()
        m.update(pickle.dumps(self.natural_key()))
        return m.digest()

    def get_thumbprint(self):
        """
        Calculates the current thumbprint of the item being tracked.
        """
        raise NotImplementedError

    def is_changed(self, last_thumbprint):
        current_thumbprint = self.get_thumbprint()
        #print('is_changed.tracker:', self)
        #print('is_changed.last_thumbprint:', last_thumbprint)
        #print('is_changed.current_thumbprint:', current_thumbprint)
        return current_thumbprint != last_thumbprint

    def act(self):
        """
        Executes the cached task. This is called by the containing satchel's configure() when `is_changed()` returns True.
        """
        if self.action:
            self.action()


class FilesystemTracker(BaseTracker):
    """
    Tracks changes to a local filesystem directory.

    Has only two custom parameters:

        base_dir = The absolute or relative local directory to search.
        extensions = A space delimited list of extension patterns to limit the search.
            These patterns are feed directly to the `find` command.
    """

    def __init__(self, base_dir='.', extensions='*.*', *args, **kwargs):
        assert os.path.isdir(base_dir), 'Directory %s does not exist.' % base_dir
        extensions = extensions.strip()
        assert extensions, 'No extensions specified.'
        super().__init__(*args, **kwargs)
        self.base_dir = base_dir
        self.extensions = extensions

    def __repr__(self):
        return '<%s %s %s>' % (type(self).__name__, self.base_dir, self.extensions)

    def natural_key(self):
        return (self.base_dir, self.extensions)

    def get_thumbprint(self):
        """
        Calculates the current thumbprint of the item being tracked.
        """
        extensions = self.extensions.split(' ')
        name_str = ' -or '.join('-name "%s"' % ext for ext in extensions)
        cmd = 'find ' + self.base_dir + r' -type f \( ' + name_str + r' \) -exec md5sum {} \; | sort -k 2 | md5sum'
        return getoutput(cmd)


class SettingsTracker(BaseTracker):
    """
    Tracks changes to one or more satchel settings.

    Has only two custom parameters:

        satchel = The satchel instance that contains the settings to track.
        names = A comma or space delimited list of setting names. If nothing specified, all names in the satchel will be tracked.
    """

    def __init__(self, satchel, names=None, *args, **kwargs):
        if isinstance(names, six.string_types):
            names = names.replace(',', ' ').split(' ')
        names = names or []
        assert isinstance(names, (tuple, list, set))
        names = sorted(set(_.strip() for _ in names if _.strip()))
        super().__init__(*args, **kwargs)
        self.satchel = satchel
        self.names = names

    @property
    def names_string(self):
        return ', '.join(self.names)

    def __repr__(self):
        return '<%s %s %s>' % (type(self).__name__, self.satchel.name, self.names_string)

    def natural_key(self):
        return (type(self.satchel).__name__, self.names_string)

    def get_thumbprint(self):
        """
        Calculates the current thumbprint of the item being tracked.
        """
        d = {}
        if self.names:
            names = self.names
        else:
            names = list(self.satchel.lenv)
        for name in self.names:
            d[name] = deepcopy(self.satchel.env[name])
        return d


class ORTracker(BaseTracker):
    """
    Computers a logical OR between two or more other trackers.
    """

    def __init__(self, *trackers, **kwargs):
        assert isinstance(trackers, (tuple, list))
        assert trackers, 'No trackers specified.'
        super().__init__(**kwargs)
        self.trackers = list(trackers)

    def __repr__(self):
        return '<%s %s>' % (type(self).__name__, self.trackers)

    def natural_key(self):
        lst = [type(self).__name__]
        for tracker in self.trackers:
            lst.extend(tracker.natural_key())
        return tuple(lst)

    def get_thumbprint(self):
        """
        Calculates the current thumbprint of the item being tracked.
        """
        d = {}
        for tracker in self.trackers:
            d[type(tracker).__name__] = tracker.get_thumbprint()
        return d



"""
Run test with:

    tox -c tox-full.ini -e py37-ubuntu_18_04_64 -- -s burlap/tests/test_apache.py::ApacheTests::test_diff

"""
import os
import shutil
import getpass

from fabric.api import settings

import burlap
from burlap.common import env, Satchel
from burlap.tests.base import TestCase
from burlap.project import project
from burlap.context import set_cwd
from burlap.deploy import STORAGE_LOCAL


class _TestSatchel(Satchel):

    name = 'test'

    def configure(self):
        pass


class ApacheTests(TestCase):

    def get_test_satchel(self):
        test = _TestSatchel()
        test.genv.hosts = ['localhost']
        test.genv.host_string = test.genv.hosts[0]
        return test

    def test_diff(self):
        """
        Confirm on a multi-site multi-host environment, apache correctly reports change.
        """

        test = self.get_test_satchel()

        print('Setting paths...')
        #env.plan_storage = STORAGE_LOCAL
        env.disable_known_hosts = True
        burlap_dir = os.path.abspath(os.path.split(burlap.__file__)[0])

        print('Initializing tmp directory...')
        d = '/tmp/test_apache_change'
        if os.path.isdir(d):
            shutil.rmtree(d)
        os.makedirs(d)

        activate_cmd = '. {d}/.env/bin/activate;'.format(d=d)
        with set_cwd(d):
            print('Creating project skeleton...')
            project.create_skeleton(
                project_name='test_apache_change',
                roles='prod',
                components='apache',
            )

            assert not os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-1' % d)
            assert not os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-2' % d)

            # Simulate multiple remote hosts by creating aliases of localhost.
            # Note, for testing this on your localhost for a user without passwordless sudo,
            # you may have to run: `sudo chmod 777 /etc/hosts`
            # This won't work on Travis, where these will instead be set in .travis.yml.
            print('Modifying /etc/hosts...')
            env.host_string = 'localhost'
            env.hosts = [env.host_string]
            env.user = getpass.getuser()
            hosts_updated = False
            with settings(warn_only=True):
                output = test.run('cat /etc/hosts')
                if 'test-dj-migrate-1' not in output:
                    for use_sudo in (False, True):
                        print('Attempting to append to /etc/hosts with use_sudo=%s...' % use_sudo)
                        #ret = append(filename='/etc/hosts', text='127.0.0.1 test-dj-migrate-1\n127.0.0.1 test-dj-migrate-2', use_sudo=use_sudo)
                        ret = test.append(filename='/etc/hosts', text='127.0.0.1 test-dj-migrate-1\n127.0.0.1 test-dj-migrate-2', use_sudo=use_sudo)
                        print('ret:', ret)
                        print('Checking /etc/hosts content...')
                        output = test.run('cat /etc/hosts')
                        print('output:', output)
                        if 'test-dj-migrate-1' in output:
                            hosts_updated = True
                            print('Success!')
                            break
                else:
                    hosts_updated = True
            assert hosts_updated

            os.system('ln -s %s %s/' % (burlap_dir, d))

            project.update_settings({
                    'plan_storage': STORAGE_LOCAL,
                    'plan_data_dir': os.path.join(d, 'plans'),
                    'services': ['apache'],
                    'default_site': 'testsite1',
                    'default_role': 'prod',
                    # This is necessary to stop get_current_hostname() from attempting to lookup our actual hostname.
                    '_ip_to_hostname': {
                        'test-dj-migrate-1': 'test-dj-migrate-1',
                        'test-dj-migrate-2': 'test-dj-migrate-2',
                    },
                    'apache_application_name': 'testsite1',
                    'apache_server_admin_email': 'sysadmin@mydomain.com',
                    'apache_server_aliases_template': '{apache_locale}.mydomain.com',
                    'apache_wsgi_dir_template': '/usr/local/{apache_application_name}/wsgi',
                    'apache_wsgi_processes': 1,
                    'apache_wsgi_threads': 0,
                },
                role='all')

            project.update_settings({
                    'hosts': ['test-dj-migrate-1', 'test-dj-migrate-2'],
                    'available_sites_by_host':{
                        'test-dj-migrate-1': [
                            'testsite1',
                        ],
                        'test-dj-migrate-2': [
                            'testsite2',
                        ]
                    },
                    'sites': {
                        'testsite1': {
                            'apache_domain_template': 'testsite1.test-dj-migrate-1.com',
                        },
                        'testsite2': {
                            'apache_domain_template': 'testsite2.test-dj-migrate-2.com',
                        },
                    },
                },
                role='prod')

            # Run a deployment preview.
            kwargs = dict(
                activate_cmd=activate_cmd,
            )
            status, output = self.getstatusoutput('{activate_cmd} fab prod deploy.preview'.format(**kwargs))
            print('output:\n%s' % output)
            assert not status
            # The deployment preview should include both hosts.
            assert "[test-dj-migrate-1] Executing task 'deploy.preview'" in output
            assert "[test-dj-migrate-2] Executing task 'deploy.preview'" in output
            assert not os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-1' % d)
            assert not os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-2' % d)

            status, output = self.getstatusoutput(
                '{activate_cmd} '
                'fab prod debug.set_satchel_value:apache,site,xyz '
                'debug.show_satchel_items:apache'.format(**kwargs))
            assert not status
            assert ' = xyz' in output

            # Fake a deployment.
            #status, output = self.getstatusoutput((
                #'{activate_cmd} '
                #'fab prod debug.set_satchel_value:apache,site,abc '
                #'deploy.fake:set_satchels="apache-site-abc"').format(**kwargs))
            #assert not status
            #assert os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-1' % d)
            #assert os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-2' % d)

            ## Confirm apache now reports no changes needing deployment.
            #status, output = self.getstatusoutput('{activate_cmd} fab prod deploy.preview'.format(**kwargs))
            #assert not status
            #assert "[test-dj-migrate-1] Executing task 'deploy.preview'" in output
            #assert "[test-dj-migrate-2] Executing task 'deploy.preview'" in output
            #assert os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-1' % d)
            #print(open('%s/plans/prod/000/thumbprints/test-dj-migrate-1' % d).read())
            #assert os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-2' % d)
            #print(open('%s/plans/prod/000/thumbprints/test-dj-migrate-2' % d).read())



"""
Supervisor processes
====================

This module provides high-level tools for managing long-running
processes using `supervisord`_.

.. _supervisord: http://supervisord.org/

"""
import os

from burlap.constants import *
from burlap import ServiceSatchel
from burlap.decorators import task


class SupervisorSatchel(ServiceSatchel):

    name = 'supervisor'

    post_deploy_command = 'restart_all'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['supervisor'],
        }

    def set_defaults(self):

        self.env.manage_configs = True
        self.env.config_template = 'supervisor/supervisor_daemon.template2.config'
        self.env.config_path = '/etc/supervisor/supervisord.conf'
        self.env.conf_dir = '/etc/supervisor/conf.d'
        self.env.daemon_bin_path_template = '{pip_virtualenv_dir}/bin/supervisord'
        self.env.daemon_path = '/etc/init.d/supervisord'
        self.env.bin_path_template = '{pip_virtualenv_dir}/bin'
        self.env.daemon_pid = '/var/run/supervisord.pid'
        self.env.log_path = "/var/log/supervisord.log"
        self.env.supervisorctl_path_template = '{pip_virtualenv_dir}/bin/supervisorctl'
        self.env.kill_pattern = ''
        self.env.services_rendered = ''

        # If true, then all configuration files not explicitly managed by use will be deleted.
        self.env.purge_all_confs = True

        self.env.services = []

        # Functions that, when called, should return a supervisor service text ready to be appended to supervisord.conf.
        # It will be called once for each site.
        self.genv._supervisor_create_service_callbacks = []

        # Most of these service commands are overridden by satchel tasks.
        # http://supervisord.org/running.html
        # http://www.onurguzel.com/supervisord-restarting-and-reloading/
        # "unix:///var/run/supervisor.sock no such file" generally indicates that the Supervisor service is not running.
        self.env.service_commands = {
            START: { # Start Supervisor service. Processes should come up automatically.
                FEDORA: 'systemctl start supervisord.service',
                UBUNTU: 'service supervisor start',
            },
            STOP: { # Stop Supervisor service. Not guaranteed to stop all processes, as stop_all() does.
                FEDORA: 'systemctl stop supervisor.service',
                UBUNTU: 'service supervisor stop',
            },
            DISABLE: {
                FEDORA: 'systemctl disable httpd.service',
                UBUNTU: 'chkconfig supervisor off',
            },
            ENABLE: {
                FEDORA: 'systemctl enable httpd.service',
                UBUNTU: 'chkconfig supervisor on',
            },
            RESTART: { # Restart Supervisor service. Not guaranteed to restart all processes, as restart_all() does.
                FEDORA: 'systemctl restart supervisord.service',
                UBUNTU: 'service supervisor restart; sleep 5',
            },
            STATUS: { # Get status of Supervisor service. Use status_all() for a better view of process status.
                FEDORA: 'systemctl --no-pager status supervisord.service',
                (UBUNTU, '14.04'): 'service supervisor status',
                (UBUNTU, '16.04'): 'systemctl --no-pager status supervisor',
                UBUNTU: 'systemctl --no-pager status supervisor',
            },
        }

    def render_paths(self):
        r = self.local_renderer
        r.env.daemon_bin_path = r.format(r.env.daemon_bin_path_template)
        r.env.bin_path = r.format(r.env.bin_path_template)
        r.env.supervisorctl_path = r.format(r.env.supervisorctl_path_template)

    def register_callback(self, f):
        self.genv._supervisor_create_service_callbacks.append(f)

    @task
    def start_all(self):
        """
        Start service and all processes.
        """
        r = self.local_renderer
        self.start()
        r.sudo('supervisorctl start all')

    @task
    def stop_all(self):
        """
        Stop all processes, but don't stop service.
        """
        r = self.local_renderer
        r.sudo('supervisorctl stop all')

    @task
    def restart_all(self):
        """
        Restart all processes, but don't restart service.
        """
        r = self.local_renderer
        r.sudo('supervisorctl restart all')

    @task
    def status_all(self):
        """
        Get status of all processes.
        """
        r = self.local_renderer
        r.sudo('supervisorctl status')

    @task
    def restart(self):
        """
        Stop all processes, reread config, and reload supervisord.
        """
        self.stop_all()
        self.reload()

    @task
    def reload(self):
        """
        Reread config and reload supervisord, but don't restart service.

        May take several minutes if processes have not yet been stopped.
        restart() is usually faster, as it uses stop_all() to stop processes.
        """
        r = self.local_renderer
        r.sudo('supervisorctl reread')
        r.sudo('supervisorctl reload')

    @task
    def update(self):
        """
        Restart applications whose configurations have changed (does not restart service).
        """
        r = self.local_renderer
        r.sudo('supervisorctl update')

    def record_manifest(self):
        """
        Called after a deployment to record any data necessary to detect changes for a future deployment.
        """
        data = super().record_manifest()

        # Celery deploys itself through supervisor, so monitor its changes too in Apache site configs.
        for site_name, site_data in self.genv.sites.items():
            if self.verbose:
                print(site_name, site_data)
            data['celery_has_worker_%s' % site_name] = site_data.get('celery_has_worker', False)

        data['configured'] = True

        # Generate services list.
        self.write_configs(upload=0)

        return data

    @task
    def write_configs(self, site=None, upload=1):

        site = site or ALL
        verbose = self.verbose

        self.render_paths()

        supervisor_services = []

        # TODO: check available_sites_by_host and remove dead?
        for _site, site_data in self.iter_sites(site=site, renderer=self.render_paths):
            if verbose:
                print('write_configs.site:', _site)
            for cb in self.genv._supervisor_create_service_callbacks:
                ret = cb(site=_site)
                if isinstance(ret, str):
                    supervisor_services.append(ret)
                elif isinstance(ret, tuple):
                    assert len(ret) == 2
                    conf_name, conf_content = ret
                    if verbose:
                        print('conf_name:', conf_name)
                        print('conf_content:', conf_content)
                    remote_fn = os.path.join(self.env.conf_dir, conf_name)
                    if int(upload):
                        local_fn = self.write_to_file(conf_content)
                        self.put(local_path=local_fn, remote_path=remote_fn, use_sudo=True)

        self.env.services_rendered = '\n'.join(supervisor_services)

        if int(upload):
            fn = self.render_to_file(self.env.config_template)
            self.put(local_path=fn, remote_path=self.env.config_path, use_sudo=True)

    def deploy_services(self, site=None, restart=1):
        """
        Collect the configurations for all registered services and write the appropriate supervisord.conf file.
        """

        verbose = self.verbose

        r = self.local_renderer
        if not r.env.manage_configs:
            return

        self.render_paths()

        supervisor_services = []

        if r.env.purge_all_confs:
            r.sudo('rm -Rf /etc/supervisor/conf.d/*')

        # TODO: check available_sites_by_host and remove dead?
        self.write_configs(site=site)
        for _site, site_data in self.iter_sites(site=site, renderer=self.render_paths):
            if verbose:
                print('deploy_services.site:', _site)

            for cb in self.genv._supervisor_create_service_callbacks:
                if self.verbose:
                    print('cb:', cb)
                ret = cb(site=_site)
                if self.verbose:
                    print('ret:', ret)
                if isinstance(ret, str):
                    supervisor_services.append(ret)
                elif isinstance(ret, tuple):
                    assert len(ret) == 2
                    conf_name, conf_content = ret
                    if self.dryrun:
                        print('supervisor conf filename:', conf_name)
                        print(conf_content)
                    self.write_to_file(conf_content)

        self.env.services_rendered = '\n'.join(supervisor_services)

        fn = self.render_to_file(self.env.config_template)
        r.put(local_path=fn, remote_path=self.env.config_path, use_sudo=True)

        if int(restart):
            self.start()
            self.update()

    @task(precursors=['packager', 'user', 'rabbitmq'])
    def configure(self, **kwargs):
        kwargs.setdefault('site', ALL)

        self.deploy_services(**kwargs)

supervisor = SupervisorSatchel()



"""
Network
=======
"""

from fabric.api import hide, run, settings, sudo

from burlap.files import file # pylint: disable=redefined-builtin

is_file = file.is_file


def interfaces():
    """
    Get the list of network interfaces. Will return all datalinks on SmartOS.
    """
    with settings(hide('running', 'stdout')):
        if is_file('/usr/sbin/dladm'):
            res = run('/usr/sbin/dladm show-link')
        else:
            res = sudo('/sbin/ifconfig -s')
    return [line.split(' ')[0] for line in res.splitlines()[1:]]


def address(interface):
    """
    Get the IPv4 address assigned to an interface.

    Example::

        import burlap

        # Print all configured IP addresses
        for interface in burlap.network.interfaces():
            print(burlap.network.address(interface))

    """
    with settings(hide('running', 'stdout')):
        res = (sudo("/sbin/ifconfig %(interface)s | grep 'inet '" % locals()) or '').split('\n')[-1].strip()
    if 'addr' in res:
        return res.split()[1].split(':')[1]
    return res.split()[1]


def nameservers():
    """
    Get the list of nameserver addresses.

    Example::

        import burlap

        # Check that all name servers are reachable
        for ip in burlap.network.nameservers():
            run('ping -c1 %s' % ip)

    """
    with settings(hide('running', 'stdout')):
        res = run(r"cat /etc/resolv.conf | grep 'nameserver' | cut -d\  -f2")
    return res.splitlines()



#!/usr/bin/python
"""
File: versioner.py

A utility for detecting the release of software library updates
from numerous sources.

Copyright (C) 2012  Chris Spencer (chrisspen at gmail dot com)

Helps you query and track software package versions.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
"""
import subprocess
import csv
import os
import re
import sys
import json

from six.moves import xmlrpc_client as xmlrpclib
from six.moves.urllib.request import urlopen

try:
    import feedparser
except ImportError:
    print('Warning: feedparser not installed', file=sys.stderr)
    feedparser = None

PIP = 'pip'
GITHUB = 'github' # Most recent commit.
GITHUB_TAG = 'github_tag' # Most recent tag.
RSS = 'rss'
APT = 'apt'
TYPES = (
    PIP,
    GITHUB,
    GITHUB_TAG,
    RSS,
    APT,
)

GITHUB_PATTERN = re.compile(
    r'https://github.com/(?P<user>[^/]+)/(?P<repo>[^/$]+)')

DEP_SCHEMA = (
    'type',
    'name',
    'uri',
    'version',
    'rss_field',
    'rss_regex',
)

def get_github_user_repo(uri):
    matches = GITHUB_PATTERN.findall(uri)
    if not matches:
        return
    user, repo = matches[0]
    return user, repo

VERSIONER_FN = os.getenv('BURLAP_VERSIONER_FN', '~/.burlap_versioner_pip')

def get_pip_oath():
    assert os.path.isfile(VERSIONER_FN), 'Credentials file %s does not exist.' % VERSIONER_FN
    client_id, client_secret = open(VERSIONER_FN, 'r').read().strip().split(',')
    return client_id, client_secret

class Dependency:

    def __init__(self, type, name, uri, version, rss_field, rss_regex): # pylint: disable=redefined-builtin
        self.type = type # source location, e.g. pip|github|rss|apt|etc
        assert type in TYPES, 'Unknown type: %s' % (self.type,)
        self.name = name
        self.uri = uri
        self.version = version
        self.rss_field = rss_field
        self.rss_regex = rss_regex
        self._cache = {}

    def __str__(self):
        return u'%s==%s' % (self.name, self.version)

    def _get_current_version_pip(self):
        client = xmlrpclib.ServerProxy('http://pypi.python.org/pypi')
        v = client.package_releases(self.uri)
        if v:
            return v[0]

    def _get_current_version_github(self):
        #TODO:use authentication to avoid rate-limiting?
        user, repo = get_github_user_repo(self.uri)
        url = 'https://api.github.com/repos/%s/%s/git/refs/heads/master' \
            % (user, repo,)
        resp = urlopen(url).read()
        resp = json.loads(resp)
        return resp['object']['sha']

    def _get_current_version_github_tag(self):
        # check rate limiting with:
        # curl -i https://api.github.com/users/whatever
        # Note, while unauthenticated, we can only access this url 60-times an hour.
        #TODO:use authentication to avoid rate-limiting?
        user, repo = get_github_user_repo(self.uri)
        url = 'https://api.github.com/repos/%s/%s/tags' % (user, repo,)
        #print 'url:',url
        resp = urlopen(url).read()
        resp = json.loads(resp)
        resp = sorted(
            resp,
            key=lambda o: tuple(int(_) if _.isdigit() else _ \
                for _ in o['name'].split('.')),
            reverse=True)
        for tag in resp:
            v = tag['name']
            if self.name in v:
                v = v.replace(self.name, '')
            if not v[0].isalnum():
                v = v[1:]
            return v

    def _get_current_version_apt(self):
#        cmd = ("apt-get upgrade --dry-run %(uri)s | grep \"Conf %(uri)s\" " + \
#            "| awk '{gsub(/[()]/,\"\"); print($3;}'") % dict(uri=self.uri)
        cmd = ("dpkg -s %(uri)s | grep -E \"Version:\" " + \
            "| awk '{gsub(/[()]/,\"\"); print($2;}'") % dict(uri=self.uri)
        #print cmd
        out = subprocess.check_output(cmd, shell=True)
        #print out
        return out

    def _get_current_version_rss(self):
        resp = feedparser.parse(self.uri)
        pat = re.compile(self.rss_regex)
        for entry in resp.entries:
            matches = pat.findall(getattr(entry, self.rss_field))
            if matches:
                return matches[0]

    def get_current_version(self):
        if '_current_version' not in self._cache:
            self._cache['_current_version'] = \
                getattr(self, '_get_current_version_%s' % self.type)()
        return self._cache['_current_version']

    def is_stale(self):
        cv = self.get_current_version()
        if not cv:
            return
        return cv != self.version

    def is_fresh(self):
        return not self.is_stale()

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(
        description='Track library dependency versions.')
    parser.add_argument('--file', dest='file', default=None,
        help='The CSV dependency file.')
    parser.add_argument('--stale', dest='stale', action='store_true',
        default=False,
        help='Only lists dependencies that have a more recent version.')
    parser.add_argument('--line', dest='line',
        help='A comma-delimited line describing a dependency.')
    parser.add_argument('--pipout', dest='pipout', action='store_true',
        default=False,
        help='If set, outputs a pip requirements file.')
    parser.add_argument('--pipin', dest='pipin',
        default=None,
        help='The filename of a pip requirements file to read and convert ' +
            'to a dependency file.')
    args = parser.parse_args()

    done = False
    if args.line:
        done = True
        dep = Dependency(*args.line.split(','))
        if dep.name == dep.uri:
            print(dep.name)
        else:
            print(dep.name, dep.uri)
        print('\tcurrent version:', dep.get_current_version())
        print('\tyour version:', dep.version)
        print('\tfresh:', dep.is_fresh())

    if args.file:
        done = True
        fn = args.file
        if not os.path.isfile(fn):
            print('Error: Dependency file %s does not exist.\n' % (fn,))
            parser.print_help()
            sys.exit(1)
        total_stale = 0
        total = 0
        for line in csv.DictReader(open(fn), delimiter=','):
            dep = Dependency(**dict((k, v) for k, v in line.items() if k))
            total += 1
            is_stale = dep.is_stale()
            total_stale += is_stale if is_stale is not None else 0
            if not args.stale or (args.stale and is_stale):
                if args.pipout:
                    if dep.type == PIP:
                        print('%s==%s' % (dep.uri, dep.version))
                    continue
                if dep.name == dep.uri:
                    print(dep.name)
                else:
                    print(dep.name, dep.uri)
                print('\tcurrent version:', dep.get_current_version())
                print('\tyour version:', dep.version)
                print('\tfresh:', dep.is_fresh())
        print('='*80)
        print('%i total dependencies' % total)
        print('%i total stale dependencies' % total_stale)
        print('%.0f%% fresh' % ((total-total_stale)/float(total)*100))

    if args.pipin:
        done = True
        fn = args.pipin
        if not os.path.isfile(fn):
            print('Error: PIP requirements file %s does not exist.\n' % (fn,))
            parser.print_help()
            sys.exit(1)
        print(','.join(DEP_SCHEMA))
        for line in open(fn, 'r').readlines():
            parts = line.strip().split('==')
            if not parts:
                continue
            args = dict(
                type='pip',
                name=parts[0],
                uri=parts[0],
                version=parts[1] if len(parts) >= 2 else '',
                rss_field='',
                rss_regex='')
            fmt = '%(type)s,%(name)s,%(uri)s,%(version)s,%(rss_field)s,' + \
                '%(rss_regex)s'
            print(fmt % args)

    if not done:
        parser.print_help()
        sys.exit(1)



from burlap.host import TimezoneSatchel
from burlap.tests.functional_tests.base import TestCase

class TimezoneTests(TestCase):

    def test_timezone(self):
        ts = TimezoneSatchel()
        ts.verbose = True

        current_tz0 = ts.get_current_timezone()
        assert current_tz0 == 'UTC'

        ts.env.timezone = 'EST'
        ts.configure()

        current_tz = ts.get_current_timezone()
        assert current_tz == 'EST'

        ts.env.timezone = current_tz0
        ts.configure()




from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class GPSDSatchel(ServiceSatchel):

    name = 'gpsd'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['gpsd', 'gpsd-clients', 'python-gps', 'ntp'],
            DEBIAN: ['gpsd', 'gpsd-clients', 'python-gps', 'ntp'],
        }

    def set_defaults(self):
        self.env.service_commands = {
            START:{
                UBUNTU: 'service gpsd start',
                DEBIAN: 'service gpsd start',
            },
            STOP:{
                UBUNTU: 'service gpsd stop',
                DEBIAN: 'service gpsd stop',
            },
            DISABLE:{
                UBUNTU: 'chkconfig gpsd off',
                (UBUNTU, '14.04'): 'update-rc.d -f gpsd remove',
                DEBIAN: 'update-rc.d gpsd disable',
            },
            ENABLE:{
                UBUNTU: 'chkconfig gpsd on',
                (UBUNTU, '14.04'): 'update-rc.d gpsd defaults',
                DEBIAN: 'update-rc.d gpsd enable',
            },
            RELOAD:{
                UBUNTU: 'service gpsd reload',
                DEBIAN: 'service gpsd reload',
            },
            RESTART:{
                UBUNTU: 'service gpsd restart; sleep 3',
                DEBIAN: 'service gpsd restart; sleep 3',
            },
        }

    @task
    def launch(self):
        self.run('gpsd /dev/ttyUSB0 -F /var/run/gpsd.sock')

    @task(precursors=['packager', 'user'])
    def configure(self):
        self.install_packages()
    configure.is_deployer = True

gpsd = GPSDSatchel()



import os

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task


class EC2MonitorSatchel(Satchel):
    """
    Wraps the EC2 monitor script provided by Amazon:

        http://docs.aws.amazon.com/AmazonCloudWatch/latest/DeveloperGuide/mon-scripts.html

    Note, the script has package dependencies described at:

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/mon-scripts.html#mon-scripts-perl_prereq
    """

    name = 'ec2monitor'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['unzip', 'libwww-perl', 'libdatetime-perl'],
        }

    def set_defaults(self):
        self.env.installed = True
        self.env.cron_path = ''
        self.env.install_path = '/home/{user}/aws-scripts-mon'
        self.env.awscreds = 'roles/{role}/aws-{role}.creds'
        self.env.awscreds_install_path = '{install_path}/aws-{role}.creds'
        self.env.options = [
            '--mem-util',
            '--disk-path=/',
            '--disk-space-util',
            '--swap-util',
            # --verify --verbose
        ]
        self.env.access_key_id = None
        self.env.secret_access_key = None
        # Run every 5 minutes by default.
        # Increase to every minute ('* * * * *') if Detailed Monitoring is enabled on the EC2 instance(s).
        self.env.schedule = '*/5 * * * *'

        self.define_cron_job(
            template='etc_crond_ec2monitor',
            script_path='/etc/cron.d/ec2monitor',
            name='default',
        )

    def _get_renderer(self, verify=False):
        r = self.local_renderer

        r.env.install_path = r.env.install_path.format(**{'user': self.genv.user})

        kwargs = {
            'role': self.genv.ROLE,
            'install_path': r.env.install_path,
        }
        r.env.awscreds = r.env.awscreds.format(**kwargs)
        r.env.awscreds_install_path = r.env.awscreds_install_path.format(**kwargs)

        options = self.env.options
        if verify:
            options.extend(['--verify --verbose'])

        r.env.command_options = ' '.join(options)
        return r

    def _get_check_command(self):
        return (
            'cd {install_path} '
            '&& export AWS_CREDENTIAL_FILE={awscreds_install_path} '
            '&& ./mon-put-instance-data.pl {command_options} '
        )

    @task
    def clear_host_data_cache(self):
        """
        Remove cached instance attribute data that may cause instance_id conflicts when an EBS volume is cloned.
        """
        r = self.local_renderer
        host_data_cache_path = '/var/tmp/aws-mon/'  # Set in CloudWatchClient::$meta_data_loc
        self.vprint(f'Removing temporary host data cache at {host_data_cache_path}.')
        r.sudo(f'rm -rf {host_data_cache_path}')

    @task
    def verify(self):
        r = self._get_renderer(verify=True)
        r.run(self._get_check_command())

    @task
    def check(self):
        r = self._get_renderer(verify=False)
        r.run(self._get_check_command())

    @task
    def install(self):
        r = self._get_renderer()

        local_path = self.env.awscreds.format(role=self.genv.ROLE)
        assert os.path.isfile(local_path), 'Missing cred file: %s' % local_path

        r.install_packages()
        r.run('cd ~; curl http://aws-cloudwatch.s3.amazonaws.com/downloads/CloudWatchMonitoringScripts-1.2.1.zip -O')
        r.run('cd ~; unzip -o CloudWatchMonitoringScripts-1.2.1.zip')
        r.run('cd ~; rm CloudWatchMonitoringScripts-1.2.1.zip')
        r.put(local_path=local_path, remote_path=r.env.awscreds_install_path)

        self.clear_host_data_cache()
        self.install_cron_job(
            name='default',
            extra={'command': self._get_check_command().format(**r.env), 'schedule': self.env.schedule}
        )

    @task
    def uninstall(self):
        self.vprint('EC2MonitorSatchel.uninstall is not yet supported.')

    @task(precursors=['packager', 'user'])
    def configure(self):
        """
        Executed when your settings have changed since the last deployment.
        Run commands to apply changes here.
        """
        self.install()


ec2monitor = EC2MonitorSatchel()


class RDSSatchel(Satchel):

    name = 'rds'

    def set_defaults(self):
        pass

    @task
    def list_instances(self):
        import boto.rds
        conn = boto.rds.connect_to_region(
            self.genv.vm_ec2_region,
            aws_access_key_id=self.genv.vm_ec2_aws_access_key_id,
            aws_secret_access_key=self.genv.vm_ec2_aws_secret_access_key,
        )
        for value in conn.get_all_dbinstances():
            print(value, value.engine, value.engine_version)

    @task(precursors=['packager', 'user'])
    def configure(self):
        pass


rds = RDSSatchel()



from fabric.api import cd

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task


class PhantomJSSatchel(Satchel):
    """
    Installs PhantomJS from upstream, since the Ubuntu package is incorrectly built.

    Extrapolated from this Gist:

        https://gist.github.com/telbiyski/ec56a92d7114b8631c906c18064ce620

    """

    name = 'phantomjs'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: [
                'build-essential',
                'chrpath',
                'libssl-dev',
                'libxft-dev',
                'libfreetype6',
                'libfreetype6-dev',
                'libfontconfig1',
                'libfontconfig1-dev',
            ],
        }

    def set_defaults(self):
        self.env.version = '2.1.1'
        self.env.target = 'phantomjs-{version}-linux-x86_64'
        self.env.wget_url = 'https://github.com/Medium/phantomjs/releases/download/v{version}/{target}.tar.bz2'
        self.env.installed = True

    @task
    def install(self):
        r = self.local_renderer

        # Ensure the buggy system package is removed.
        r.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq purge phantomjs')

        with cd('/tmp'):
            r.run('wget {wget_url}')
            r.run('tar xvjf {target}.tar.bz2')
            r.sudo('[ -d /usr/local/share/{target} ] && rm -Rf /usr/local/share/{target} || true')
            r.sudo('mv {target} /usr/local/share')
            r.sudo('[ -f /usr/local/bin/phantomjs ] && rm -f /usr/local/bin/phantomjs || true')
            r.sudo('ln -sf /usr/local/share/{target}/bin/phantomjs /usr/local/bin')
            r.run('phantomjs --version')

    @task
    def uninstall(self):
        r = self.local_renderer
        r.sudo('rm -Rf /usr/local/share/{target}')
        r.sudo('rm -f /usr/local/bin/phantomjs')

    @task(precursors=['packager', 'user'])
    def configure(self, *args, **kwargs):
        r = self.local_renderer
        if r.env.installed:
            self.install()
        else:
            self.uninstall()

phantomjs = PhantomJSSatchel()



import os
from pprint import pprint

from burlap.common import Satchel, is_local
from burlap.constants import *
from burlap.decorators import task

EZ_SETUP = 'ez_setup'
PYTHON_PIP = 'python-pip'
GET_PIP = 'get-pip'
BOOTSTRAP_METHODS = (
    EZ_SETUP,
    PYTHON_PIP,
    GET_PIP,
)


class PIPSatchel(Satchel):

    name = 'pip'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: [
                'gcc', 'python3-dev', 'build-essential', 'python3-pip',
            ],
        }

    def set_defaults(self):
        self.env.bootstrap_method = PYTHON_PIP
        self.env.user = 'www-data'
        self.env.group = 'www-data'
        self.env.perms = '775'
        self.env.virtualenv_dir = '.env'
        self.env.requirements = 'pip-requirements.txt'
        self.env.python_version = '3.7'

    @task
    def has_pip(self):
        """
        Is pip3 installed on the system?
        """
        r = self.local_renderer
        with self.settings(warn_only=True):
            ret = (r.run('which pip3') or '').strip()
            if ret:
                print(f'Pip is installed at {ret}.')
            else:
                print('Pip is not installed.')
            return bool(ret)

    @task
    def bootstrap(self, force=0):
        """
        Install all the packages necessary for managing virtual environments with pip.
        """
        force = int(force)
        if self.has_pip() and not force:
            return

        r = self.local_renderer

        if r.env.bootstrap_method == GET_PIP:
            r.sudo('curl --silent --show-error --retry 5 https://bootstrap.pypa.io/get-pip.py | python')
        elif r.env.bootstrap_method == EZ_SETUP:
            r.run('wget http://peak.telecommunity.com/dist/ez_setup.py -O /tmp/ez_setup.py')
            with self.settings(warn_only=True):
                r.sudo('python /tmp/ez_setup.py -U setuptools')
            r.sudo('easy_install -U pip')
        elif r.env.bootstrap_method == PYTHON_PIP:
            r.sudo('apt-get install -y python3-pip')
        else:
            raise NotImplementedError('Unknown pip bootstrap method: %s' % r.env.bootstrap_method)

        # Upgrade pip
        r.run_or_local(f'python{self.env.python_version} -m pip install pip')

    @task
    def clean(self, virtualenv_dir=None):
        r = self.local_renderer
        with self.settings(warn_only=True):
            print('Deleting old virtual environment...')
            r.sudo('rm -Rf {virtualenv_dir}')

    @task
    def clean_virtualenv(self, *args, **kwargs):
        self.clean(*args, **kwargs)

    @task
    def virtualenv_exists(self, virtualenv_dir=None):
        """
        Return True if the virtual environment has been created.
        """
        r = self.local_renderer
        if not virtualenv_dir:
            virtualenv_dir = self.env.virtualenv_dir
        with self.settings(warn_only=True):
            ls_output = r.run_or_local('ls {virtualenv_dir}') or ''
            ret = 'bin' in ls_output.strip()  # Virtualenv should have a 'bin' directory

        if ret:
            self.vprint(f'Virtualenv exists at {virtualenv_dir}.')
        else:
            self.vprint(f'Virtualenv does not exist at {virtualenv_dir}.')

        return ret

    @task
    def what_requires(self, name):
        """
        List the packages that require the given package.
        """
        r = self.local_renderer
        r.env.name = name
        r.local('pipdeptree -p {name} --reverse')

    @task
    def create_virtualenv(self):
        """
        Create the virtual environment if it doesn't exist already.
        """
        r = self.local_renderer
        if not self.virtualenv_exists():
            print('Creating new virtual environment...')
            print('shell_env:', self.genv.shell_env)
            if is_local():
                r.local('python{python_version} -m venv {virtualenv_dir}')
            else:
                r.sudo('python{python_version} -m venv {virtualenv_dir}', user=r.env.user)

    def get_combined_requirements(self, requirements=None):
        """
        Returns all requirements files combined into one string.
        """

        def iter_lines(fn):
            with open(fn, 'r') as fin:
                for line in fin.readlines():
                    line = line.strip()
                    if not line or line.startswith('#'):
                        continue
                    if line.startswith('-r'):
                        recursive_requirements = line[3:]
                        self.vprint('Recursively including requirements from {}.'.format(recursive_requirements))
                        recursive_requirements_path = os.path.join(
                            self.genv.ROLES_DIR, self.genv.ROLE, recursive_requirements
                        )
                        yield from iter_lines(recursive_requirements_path)
                    else:
                        yield line

        requirements = requirements or self.env.requirements
        if isinstance(requirements, str):
            requirements = [requirements]

        content = []
        for requirement in requirements:
            requirements_path = self.find_template(requirement)
            content.extend(list(iter_lines(requirements_path)))

        return '\n'.join(content)

    @task
    def update_install(self, **kwargs):
        r = self.local_renderer
        r.env.quiet_flag = '' if self.verbose else '-q'

        options = [
            'requirements',
            'virtualenv_dir',
            'user',
            'group',
            'perms',
        ]
        for option in options:
            setattr(r.env, option, kwargs.pop(option, None) or getattr(r.env, option))

        # Make sure pip is installed.
        self.bootstrap()

        # Make sure our virtualenv is installed.
        self.create_virtualenv()

        # Collect all requirements.
        tmp_fn = r.write_temp_file(self.get_combined_requirements(requirements=r.env.requirements))

        # Copy up our requirements.
        r.env.pip_remote_requirements_fn = '/tmp/pip-requirements.txt'
        r.put(local_path=tmp_fn, remote_path=r.env.pip_remote_requirements_fn)

        # Ensure we're always using the latest pip.
        if is_local():
            r.local('{virtualenv_dir}/bin/pip {quiet_flag} install -U pip')
        else:
            r.sudo('{virtualenv_dir}/bin/pip {quiet_flag} install -U pip', user=r.env.user)

        # Install requirements from file.
        if is_local():
            r.local("{virtualenv_dir}/bin/pip {quiet_flag} install --exists-action w -r {pip_remote_requirements_fn}")
        else:
            r.sudo("{virtualenv_dir}/bin/pip {quiet_flag} install --exists-action w -r {pip_remote_requirements_fn}", user=r.env.user)

    @task
    def record_manifest(self):
        """
        Called after a deployment to record any data necessary to detect changes for a future deployment.
        """
        manifest = super().record_manifest()
        manifest['all-requirements'] = self.get_combined_requirements()
        if self.verbose:
            pprint(manifest, indent=4)
        return manifest

    @task(precursors=['packager', 'user'])
    def configure(self, *args, **kwargs):
        clean = int(kwargs.pop('clean', 0))
        if clean:
            self.clean_virtualenv()

        # Necessary to make warning message go away.
        # http://stackoverflow.com/q/27870003/247542
        self.genv['sudo_prefix'] += '-H '

        self.update_install(*args, **kwargs)

pip = PIPSatchel()



"""
Helper functions for sending a notification email before and after each deployment, and on deployment failures.
"""
import smtplib
import traceback
from email.mime.text import MIMEText

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task


class DeploymentNotifierSatchel(Satchel):

    name = 'deploymentnotifier'

    def set_defaults(self):

        self.env.email_enabled = False
        self.env.email_host = None
        self.env.email_port = 587
        self.env.email_host_user = None
        self.env.email_host_password = None
        self.env.email_use_tls = True
        self.env.email_recipient_list = []

    def send_email(self, subject, message, from_email=None, recipient_list=None):
        recipient_list = recipient_list or []
        if not recipient_list:
            self.vprint('send_email() aborted: No recipient_list.')
            return

        if self.dryrun:
            self.print_command('echo -e "{body}" | mail -s "{subject}" {recipients}'.format(
                recipients=','.join(self.env.email_recipient_list),
                body=message.replace('\n', '\\n'),
                subject=subject,
            ))
            return

        from_email = from_email or self.env.email_host_user

        msg = MIMEText(message)

        # me == the sender's email address
        # you == the recipient's email address
        msg['Subject'] = subject
        msg['From'] = from_email
        msg['To'] = '; '.join(recipient_list)

        # Send the message via our own SMTP server, but don't include the
        # envelope header.
        print('Attempting to send mail using %s@%s...' \
            % (self.env.email_host_user, self.env.email_host))
        s = smtplib.SMTP(self.env.email_host, self.env.email_port)
        s.ehlo()
        s.starttls()
        s.ehlo()
        self.vprint('user:', self.env.email_host_user, 'password:', self.env.email_host_password)
        s.login(self.env.email_host_user, self.env.email_host_password)
        s.sendmail(from_email, recipient_list, msg.as_string())
        s.quit()

    @task
    def test(self):
        self.send_email(
            subject='Test',
            message='Test',
            recipient_list=self.env.email_recipient_list)

    @task
    def notify_pre_deployment(self, subject=None, message=None, force=0):
        """
        Send email notifying recipients of deploy start.
        """
        force = int(force)

        if force or self.env.email_enabled and self.genv.host_string == self.genv.hosts[0]: # First host only
            subject = subject or '{} Deployment Started'.format(self.genv.ROLE.title())
            message = message or 'Deployment to {} has started.'.format(self.genv.ROLE)
            self.send_email(subject=subject, message=message, recipient_list=self.env.email_recipient_list)
        else:
            self.vprint('Skipping notify_pre_deployment.')

    @task
    def notify_post_deployment(self, subject=None, message=None, force=0):
        """
        Send email notifying recipients of deploy completion.
        """
        force = int(force)

        if force or self.env.email_enabled and self.genv.host_string == self.genv.hosts[-1]: # Last host only
            subject = subject or '{} Deployment Complete'.format(self.genv.ROLE.title())
            message = message or 'Deployment to {} is complete.'.format(self.genv.ROLE)
            self.send_email(subject=subject, message=message, recipient_list=self.env.email_recipient_list)
        else:
            self.vprint('Skipping notify_post_deployment.')

    @task
    def notify_failed_deployment(self, subject=None, message=None, force=0):
        """
        Send email notifying recipients of deploy failure, including traceback if applicable.
        """
        force = int(force)

        if force or self.env.email_enabled:
            subject = subject or '{} Deployment Failed'.format(self.genv.ROLE.title())
            message = message or 'Deployment to {role} has failed.\n\n{tb}'.format(
                role=self.genv.ROLE, tb=traceback.format_exc()
            )
            self.send_email(subject=subject, message=message, recipient_list=self.env.email_recipient_list)
        else:
            self.vprint('Skipping notify_failed_deployment.')

    @task
    def configure(self):
        pass


class LoginNotifierSatchel(Satchel):
    """
    Causes a notification email to be sent whenever someone logs into the server.

    The mail functionality setup by this satchel is pretty light, and generally assumes your server hasn't been blacklisted for spamming.

    If you require a specific SMTP login to send email, then you might have to enable the postfix satchel as well.
    """

    name = 'loginnotifier'

    def set_defaults(self):
        self.env.sysadmin_email = 'root@localhost'
        self.env.script_template = 'notifier/loginnotifier.template.sh'
        self.env.script_installation_path = '/etc/profile.d/loginnotifier.sh'
        self.env.script_user = 'root'
        self.env.script_group = 'root'
        self.env.script_chmod = 'u+rx,g+rx'
        self.env.append_bash_aliases = False

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: [
                # 'postfix',
                'mailutils',
                # 'libsasl2-2',
                # 'ca-certificates',
                # 'libsasl2-modules',
                # 'nano',
            ],
        }

    @task(precursors=['packager', 'user'])
    def configure(self):
        r = self.local_renderer
        if self.env.enabled:
            fn = self.render_to_file(self.env.script_template)
            r.put(
                local_path=fn,
                remote_path=self.env.script_installation_path, use_sudo=True)
            r.sudo('chown {script_user}:{script_group} {script_installation_path}')
            r.sudo('chmod {script_chmod} {script_installation_path}')
        else:
            r.sudo('rm {script_installation_path}')

        if r.env.append_bash_aliases:
            r.append(text='bash %s' % r.env.script_installation_path, filename='~/.bash_aliases')


deployment_notifier = DeploymentNotifierSatchel()
notify_pre_deployment = deployment_notifier.notify_pre_deployment
notify_post_deployment = deployment_notifier.notify_post_deployment
notify_failed_deployment = deployment_notifier.notify_failed_deployment
send_email = deployment_notifier.send_email

login_notifier = LoginNotifierSatchel()



import re

from fabric.api import hide

from six.moves.urllib.request import urlopen

from burlap.constants import *
from burlap import ContainerSatchel
from burlap.decorators import task

def _to_int(val):
    try:
        return int(val)
    except ValueError:
        return val

DOWNLOAD_LINK_PATTERN = re.compile(r'http[s]{0,1}://[^/]+/vagrant/[0-9\.]+/vagrant[^"]+')

class VagrantSatchel(ContainerSatchel):

    name = 'vagrant'

    def set_defaults(self):
        self.env.box = '?'
        self.env.provider = '?'
        self.env.shell_command = 'vagrant ssh'
        self.env.download_url = 'https://www.vagrantup.com/downloads.html'

    def ssh_config(self, name=''):
        """
        Get the SSH parameters for connecting to a vagrant VM.
        """
        r = self.local_renderer
        with self.settings(hide('running')):
            output = r.local('vagrant ssh-config %s' % name, capture=True)

        config = {}
        for line in output.splitlines()[1:]:
            key, value = line.strip().split(' ', 2)
            config[key] = value
        return config

    def _get_settings(self, config):
        settings = {}

        user = config['User']
        hostname = config['HostName']
        port = config['Port']

        # Build host string
        host_string = "%s@%s:%s" % (user, hostname, port)

        settings['user'] = user
        settings['hosts'] = [host_string]
        settings['host_string'] = host_string

        # Strip leading and trailing double quotes introduced by vagrant 1.1
        settings['key_filename'] = config['IdentityFile'].strip('"')

        settings['forward_agent'] = (config.get('ForwardAgent', 'no') == 'yes')
        settings['disable_known_hosts'] = True

        return settings


    @task
    def version(self):
        """
        Get the Vagrant version.
        """
        r = self.local_renderer
        with self.settings(hide('running', 'warnings'), warn_only=True):
            res = r.local('vagrant --version', capture=True)
        if res.failed:
            return None
        line = res.splitlines()[-1]
        version = re.match(r'Vagrant (?:v(?:ersion )?)?(.*)', line).group(1)
        return tuple(_to_int(part) for part in version.split('.'))


    @task
    def setup(self, name=''):
        r = self.local_renderer
        _settings = self._get_settings(self.ssh_config(name=name))
        if self.verbose:
            print(_settings)
        r.genv.update(_settings)


    @task
    def init(self):
        r = self.local_renderer
        r.local('vagrant init {box}')


    @task
    def up(self):
        r = self.local_renderer
        r.local('vagrant up --provider={provider}')


    @task
    def shell(self):
        r = self.local_renderer
        self.setup()
        r.local(r.env.shell_command)


    @task
    def destroy(self):
        r = self.local_renderer
        r.local('vagrant destroy')


    @task
    def upload(self, src, dst=None):
        r = self.local_renderer
        r.put(local_path=src, remote_path=dst)


    #http://serverfault.com/a/758017/41252
    @task
    def ssh(self):
        r = self.local_renderer
        self.setup()
        hostname, port = r.genv.host_string.split('@')[-1].split(':')
        r.local((
            'ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no '
            '-i %s %s@%s -p %s') % (r.genv.key_filename, r.genv.user, hostname, port))


    def _settings_dict(self, config):
        settings = {}

        user = config['User']
        hostname = config['HostName']
        port = config['Port']

        # Build host string
        host_string = "%s@%s:%s" % (user, hostname, port)

        settings['user'] = user
        settings['hosts'] = [host_string]
        settings['host_string'] = host_string

        # Strip leading and trailing double quotes introduced by vagrant 1.1
        settings['key_filename'] = config['IdentityFile'].strip('"')

        settings['forward_agent'] = (config.get('ForwardAgent', 'no') == 'yes')
        settings['disable_known_hosts'] = True

        return settings


    @task
    def vagrant(self, name=''):
        """
        Run the following tasks on a vagrant box.

        First, you need to import this task in your ``fabfile.py``::

            from fabric.api import *
            from burlap.vagrant import vagrant

            @task
            def some_task():
                run('echo hello')

        Then you can easily run tasks on your current Vagrant box::

            $ fab vagrant some_task

        """
        r = self.local_renderer
        config = self.ssh_config(name)

        extra_args = self._settings_dict(config)
        r.genv.update(extra_args)

    def vagrant_settings(self, name='', *args, **kwargs):
        """
        Context manager that sets a vagrant VM
        as the remote host.

        Use this context manager inside a task to run commands
        on your current Vagrant box::

            from burlap.vagrant import vagrant_settings

            with vagrant_settings():
                run('hostname')
        """
        config = self.ssh_config(name)

        extra_args = self._settings_dict(config)
        kwargs.update(extra_args)

        return self.settings(*args, **kwargs)

    def status(self, name='default'):
        """
        Get the status of a vagrant machine
        """
        machine_states = dict(self._status())
        return machine_states[name]

    def _status(self):
        if self.version() >= (1, 4):
            return self._status_machine_readable()
        return self._status_human_readable()

    def _status_machine_readable(self):
        with self.settings(hide('running')):
            output = self.local('vagrant status --machine-readable', capture=True)
        tuples = [tuple(line.split(',')) for line in output.splitlines() if line.strip() != '']
        return [(target, data) for timestamp, target, type_, data in tuples if type_ == 'state-human-short']

    def _status_human_readable(self):
        with self.settings(hide('running')):
            output = self.local('vagrant status', capture=True)
        lines = output.splitlines()[2:]
        states = []
        for line in lines:
            if line == '':
                break
            target = line[:25].strip()
            state = re.match(r'(.{25}) ([^\(]+)( \(.+\))?$', line).group(2)
            states.append((target, state))
        return states

    def machines(self):
        """
        Get the list of vagrant machines
        """
        return [name for name, state in self._status()]

    def base_boxes(self):
        """
        Get the list of vagrant base boxes
        """
        return sorted(list(set(name for name, provider in self._box_list())))

    def _box_list(self):
        if self.version() >= (1, 4):
            return self._box_list_machine_readable()
        return self._box_list_human_readable()

    def _box_list_machine_readable(self):
        r = self.local_renderer
        with self.settings(hide('running')):
            output = r.local('vagrant box list --machine-readable', capture=True)
        tuples = [tuple(line.split(',')) for line in output.splitlines() if line.strip() != '']
        res = []
        for timestamp, target, type_, data in tuples:
            if type_ == 'box-name':
                box_name = data
            elif type_ == 'box-provider':
                box_provider = data
                res.append((box_name, box_provider))
            else:
                raise ValueError('Unknown item type')
        return res

    def _box_list_human_readable(self):
        r = self.local_renderer
        with self.settings(hide('running')):
            output = r.local('vagrant box list', capture=True)
        lines = output.splitlines()
        res = []
        for line in lines:
            box_name = line[:25].strip()
            mo = re.match(r'.{25} \((.+)\)$', line)
            box_provider = mo.group(1) if mo is not None else 'virtualbox'
            res.append((box_name, box_provider))
        return res

    @task
    def install_from_upstream(self):
        """
        Installs Vagrant from the most recent package available from their homepage.
        """
        from burlap.system import get_arch, distrib_family
        r = self.local_renderer
        content = urlopen(r.env.download_url).read()
        print(len(content))
        matches = DOWNLOAD_LINK_PATTERN.findall(content)
        print(matches)
        arch = get_arch() # e.g. 'x86_64'
        family = distrib_family()
        if family == DEBIAN:
            ext = '.deb'
            matches = [match for match in matches if match.endswith(ext) and arch in match]
            print('matches:', matches)
            assert matches, "No matches found."
            assert len(matches) == 1, "Too many matches found: %s" % (', '.join(matches))
            r.env.final_download_url = matches[0]
            r.env.local_filename = '/tmp/vagrant%s' % ext
            r.run('wget -O {local_filename} {final_download_url}')
            r.sudo('dpkg -i {local_filename}')
        else:
            raise NotImplementedError('Unsupported family: %s' % family)

    @task(precursors=['packager', 'user'])
    def configure(self, *args, **kwargs):
        self.install_from_upstream()

vagrant = VagrantSatchel()
